import random
from os import path
from click import echo, secho, get_current_context
from collections import namedtuple

PACKAGE_DIR = path.dirname(__file__)

def chunker(seq, size):
    return [seq[pos:pos + size] for pos in range(0, len(seq), size)]

def _force_color_in_workers():
    try:
        from llacie import app
        ctx = get_current_context()
        current_app = ctx.find_object(app.App)
        if current_app is None: return None
        return current_app.is_worker or None
    except RuntimeError:
        pass
    return None

def echo_err(message):
    secho(message, fg='red', err=True, color=_force_color_in_workers())

def echo_warn(message):
    secho(message, fg='yellow', err=True, color=_force_color_in_workers())

def echo_info(message):
    secho(message, fg='green', err=True, color=_force_color_in_workers())

def strtobool(val):
    """Convert a string representation of truth to true (1) or false (0).
    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false values
    are 'n', 'no', 'f', 'false', 'off', and '0'.  Raises ValueError if
    'val' is anything else.
    """
    if isinstance(val, bool): return val
    val = val.lower()
    if val in ('y', 'yes', 't', 'true', 'on', '1'):
        return True
    elif val in ('n', 'no', 'f', 'false', 'off', '0'):
        return False
    else:
        raise ValueError("invalid truth value %r" % (val,))

def create_row_like(**kwargs):
    """Create a row-like object with positional and attribute access"""
    RowLike = namedtuple('RowLike', kwargs.keys())
    return RowLike(**kwargs)

def generate_unique_id(existing_ids, digits=10):
    """Generate a large random integer not in `existing_ids`. Each integer is `digits` long.
    Returns the new ID and also adds it to `existing_ids`."""
    while True:
        new_id = random.randint(10 ** digits, 10 ** (digits + 1) - 1)
        if new_id not in existing_ids:
            existing_ids.add(new_id)
            return new_id

def echo_strategies(strategies, title="Sections"):
    last_task_name = None
    echo(f"\n{(title + ':').ljust(24)} Available strategies:")
    echo("════════════════════════ ════════════════════════════════════════")
    for i, strat in enumerate(strategies):
        next_task_name = strategies[i + 1].task.name if i + 1 < len(strategies) else None
        if strat.task.name == last_task_name:
            sep = "  └─" if next_task_name != strat.task.name else "  ├─"
            task_formatted = (" " * 20)
        else: 
            sep = "◁───" if next_task_name != strat.task.name else "◁─┬─"
            task_formatted = strat.task.name.ljust(20)
        echo(f"    {task_formatted} {sep} {strat.name}")
        last_task_name = strat.task.name
    echo("")

def only_one_strategy(strategies, title="Sections"):
    if len(strategies) > 1:
        echo("Error: Your parameters matched several strategies. Here's what matched:")
        echo_strategies(strategies, title="Sections")
        echo("Use the --strategy-name option to select a single strategy.")
        return False
    elif len(strategies) == 0:
        echo("Error: No strategies matching those parameters were found.")
        return False
    return True
    


import os
import pandas as pd

from click import echo
from sqlalchemy import create_engine, text
from tqdm import tqdm

from .config import ConfigError
from .utils import chunker

class EDWDatabase(object):
    ADMIT_INPATIENT_NOTE_TYPE_CDS = (4, 6, 8, 19, 26, 1000010)
    # H&P, ED Notes, ED Triage Notes, ED Provider Notes, H&P (View-Only), 
    # ED Progress/Update Note
    CONSULT_INPATIENT_NOTE_TYPE_CDS = (2,)
    # Consult Notes
    PROGRESS_INPATIENT_NOTE_TYPE_CDS = (1,)
    # Progress Notes
    DISCHARGE_INPATIENT_NOTE_TYPE_CDS = (5,)
    # Discharge Summary

    def __init__(self, config):
        if config['EDW_URI'] is None:
            raise ConfigError("Could not build EDW connection string. Please"
                " check that EDW_HOST, EDW_USER, and EDW_PASS are configured"
                " in .env, .env.example, and/or ~/.edw_env.")
        
        override_odbcsysini = config['EDW_OVERRIDE_ODBCSYSINI']
        if override_odbcsysini != '' and override_odbcsysini is not None:
            if override_odbcsysini[0] == '/':
                os.environ['ODBCSYSINI'] = override_odbcsysini
            else:
                os.environ['ODBCSYSINI'] = os.path.abspath(os.path.join(__file__,
                    "..", override_odbcsysini))

        self.conn = create_engine(config['EDW_URI']).connect()
        self.config = config


    def test_query(self):
        test = self.conn.execute(
            text("SELECT TOP 3 * FROM Epic.Reference.Department AS dep"))
        echo(test.fetchall())
    

    def gen_note_metadata(self, all_ec_ids, inpt_note_type_cds, chunk_size=1000):
        """Generator for note metadata given an iterable of `PatientEncounterID`s
        and an iterable of `InpatientNoteTypeCD`s.
        
        Yields pandas `DataFrame`s with up to `chunk_size` rows each."""

        desc = f"In batches of {chunk_size} rows"
        for ec_ids in tqdm(chunker(all_ec_ids, chunk_size), desc=desc):
            select_notes_sql = f"""
                SELECT ecn.NoteID,
                        ecn.PatientEncounterID,
                        ecn.InpatientNoteTypeDSC,
                        ecn.CurrentAuthorID,
                        ecnei.AuthorServiceDSC,
                        ecnei.AuthorProviderTypeDSC,
                        ecn.CreateDTS,
                        ecn.DateOfServiceDTS,
                        ecn.LastFiledDTS,
                        ecn.EDWLastModifiedDTS,
                        ecnei.NoteFiledDTS
                    FROM Epic.Clinical.Note_PHS AS ecn
                    LEFT JOIN Epic.Clinical.NoteEncounterInformation_PHS AS ecnei
                        ON ecn.NoteID = ecnei.NoteID
                    WHERE ecn.PatientEncounterID IN (
                            {','.join([str(ec_id) for ec_id in ec_ids])})
                        AND ecn.InpatientNoteTypeCD IN (
                            {','.join([str(code) for code in inpt_note_type_cds])})
                        AND ecn.AmbulatoryNoteFLG = 'N'
                        AND (ecn.DeletedStatusCategoryCD IS NULL 
                            OR ecn.DeletedStatusCategoryCD != 2)
                        AND (ecn.UnsignedFLG IS NULL OR ecn.UnsignedFLG != 'Y')
                        AND ecnei.MostRecentContactFLG = 'Y'
                    ORDER BY ecn.CreateDTS
                """
            df = pd.read_sql_query(text(select_notes_sql), self.conn)
            df_sorted = df.sort_values(by=['NoteID', 'NoteFiledDTS'])
            df_dedup = df_sorted.drop_duplicates(subset='NoteID', keep='last').copy()
            yield df_dedup

    def gen_admission_enc_note_metadata(self, adm_ec_ids, *args, **kwargs):
        note_type_codes = (self.ADMIT_INPATIENT_NOTE_TYPE_CDS +
            self.CONSULT_INPATIENT_NOTE_TYPE_CDS + self.PROGRESS_INPATIENT_NOTE_TYPE_CDS)
        return self.gen_note_metadata(adm_ec_ids, note_type_codes, *args, **kwargs)

    def gen_discharge_enc_note_metadata(self, dc_ec_ids, *args, **kwargs):
        return self.gen_note_metadata(dc_ec_ids,
            self.DISCHARGE_INPATIENT_NOTE_TYPE_CDS, *args, **kwargs)


    def gen_note_text(self, all_note_ids, chunk_size=1000):
        """Generator for note full text given an iterable of `NoteID`s.
        
        Yields pandas `DataFrame`s with up to `chunk_size` rows each."""

        desc = f"In batches of {chunk_size} rows"
        for note_ids in tqdm(chunker(all_note_ids, chunk_size), desc=desc):
            # Get full note texts for the latest versions of each note
            note_id_sql = "','".join([str(note_id) for note_id in note_ids])
            select_note_text_sql = f"""
                SELECT NoteCSNID, ContactDateRealNBR, NoteID, LineNBR, NoteTXT
                    FROM (
                        SELECT NoteCSNID, ContactDateRealNBR, NoteID, LineNBR, NoteTXT,
                                MAX(ContactDateRealNBR) OVER (PARTITION BY NoteID)
                                    AS max_contact_date_for_note_id
                            FROM Epic.Clinical.NoteText_PHS
                            WHERE NoteID IN ('{note_id_sql}')
                        ) AS ecnt
                    WHERE max_contact_date_for_note_id = ContactDateRealNBR
                    ORDER BY NoteID ASC, LineNBR ASC
                """
            df = pd.read_sql_query(text(select_note_text_sql), self.conn)

            # Sometimes there are zero rows, for a chunk of NoteIDs with no NoteTXT
            if len(df) == 0: continue
            # Join the text from multiple lines/rows into a single field
            def join_func(pieces):
                return "\n".join([(p if p is not None else "") for p in pieces])
            df['note_text'] = df.groupby(['NoteID'])['NoteTXT'].transform(join_func)
            # Yield a selection of columns, after deduplicating
            df = df[['NoteCSNID', 'ContactDateRealNBR', 'NoteID', 'note_text']]
            yield df.drop_duplicates()
    
    
    def close(self):
        self.conn.close()


import subprocess as sub
import shlex
import time

from os import path
from textwrap import dedent

from .slurm import SlurmJobManager
from ..utils import chunker, echo_warn, echo_info, echo_err
from ..cache.abstract import AbstractFeatureWorkerCache

class ApptainerSlurmJobManager(SlurmJobManager):
    """
    Uses slurm in the specific manner expected by ERISXdl after its rebuild in July 2025, where 
    containers are launched using Apptainer instead of podman and Kubernetes.

    There no longer is a wrapper script to use Kubernetes to deploy podman containers. Per the email
    from HPC support on 7/10/2025, "Kubernetes was removed from the platform hence there is no 
    requirement to specify a wrapper script at the end of your slurm job script. Basically, slurm 
    jobs are submitted in the same way as for eristwo." Use of Apptainer on ERIS is documented at:
    https://rc.partners.org/kb/article/3888

    Apptainer is the new name for Singularity, a project that adapted containerization 
    techniques popularized by Docker to the needs of users in scientific high-performance computing
    environments. For instance, containers are launched without root privileges, no daemon process
    running as root is required, container images are stored in self-contained ".sif" files, and GPUs 
    are natively supported. This all suits low-privilege users sharing nodes with a job scheduler:
    https://journals.plos.org/plosone/article?id=10.1371/journal.pone.0177459

    Apptainer images can be created from Docker images using the following techniques:
    https://rc.partners.org/kb/computational-resources/linux-cluster/applications?article=4119
    """

    # The smallest walltime limit for all paid queues is 2 hours.
    # At about ~1 sec per ID, a 2000-ID job takes ~40min
    SLURM_IDS_PER_JOB = 2000
    SLURM_PARTITION = "Short,Medium,Long"
    SLURM_VENV = '.venv-xdl'  # <-- relative to the root of the repo
    SLURM_THREADS_PER_JOB = 10

    # Configuration that is passed into the wrapper script
    APPTAINER_IMAGE = '/data/tide/trp10/singularity/cuda-vllm-llacie.sif' # Singularity image file
    APPTAINER_DATA_VOLUME = '/data/tide'

    # Currently, ERISXdl GPU nodes don't have any outbound networking at all, not even allowing
    # connections to the Postgres database. As a workaround, we use a filesystem-based cache as an 
    # intermediary. This WorkerCache functionality is implemented in the llacie.cache.* modules
    APPTAINER_WORKERS_HAVE_NETWORKING = False
    WORKER_CACHE_POLL_INTERVAL = 15

    # Quick refresher on common options for Slurm's sbatch and srun:
    # -p = Slurm partition (aka queue)
    # -n, --ntasks = how many processes
    #      (note that the number of cores per process is set in `_create_slurm_job()` with -c)
    # -N = how many nodes (min-max or one number)
    # -t = time limit in HH:MM:SS
    # --mem = total memory limit, suffixed in K/M/G/T
    # -x = exclude certain nodes by name (e.g., if they are misbehaving)
    SLURM_OPTS = "-n 1 -t 2:00:00 -x dgx-3 --gpus=1 --mem=16G"

    SLURM_JOB_TEMPLATE = dedent("""\
        #!/bin/bash
        ## Set the docker container image to be used in the job runtime.
        export APPTAINER_IMAGE={apptainer_image}

        ## Parameters for the script that the container will run
        ids_file={ids_file}
        cache_file={cache_file}
        venv_activate={venv_activate}
        export APPTAINER_SCRIPT_VARS="$venv_activate $cache_file $ids_file {run_worker}"

        # Define group briefcase
        export APPTAINER_DATA_VOLUME={apptainer_data_volume}
        export SINGULARITY_BIND="$APPTAINER_DATA_VOLUME:$APPTAINER_DATA_VOLUME"

        ## Runs the Apptainer container as described at https://rc.partners.org/kb/article/3888
        module load singularity/3.7.0
        srun singularity exec --nv "$APPTAINER_IMAGE" {apptainer_script}
        """)


    def __init__(self, worker_cmd, config=dict(), worker_cache=None):
        super().__init__(worker_cmd, config)
        self.threads = int(config.get("SLURM_THREADS_PER_JOB", self.SLURM_THREADS_PER_JOB))

        self.apptainer_image = config.get("APPTAINER_IMAGE", self.APPTAINER_IMAGE)
        self.apptainer_data_volume = config.get("APPTAINER_DATA_VOLUME", self.APPTAINER_DATA_VOLUME)

        # This JobManager does not ever support interactive jobs
        self.slurm_interactive = False
        self.array_job_id = None
        self._done_counter = 0

        # The behavior of this JobManager changes if the workers don't have any network capability
        # In this case, we use a filesystem-based cache
        self.workers_have_networking = config.get("APPTAINER_WORKERS_HAVE_NETWORKING", 
            self.APPTAINER_WORKERS_HAVE_NETWORKING)
        self.worker_cache = worker_cache
        if not self.workers_have_networking:
            if not isinstance(self.worker_cache, AbstractFeatureWorkerCache):
                raise RuntimeError("When providing a worker cache to a ApptainerSlurmJobManager, "
                    "it must inherit from AbstractFeatureWorkerCache")
            if self.worker_cache is None:
                raise RuntimeError("ApptainerSlurmJobManager requires a worker cache if the "
                    "workers do not have networking enabled!")


    # We override the creation of the sbatch script since it is more complicated with the 
    # introduction of the wrapper script.
    def _create_sbatch_script(self):
        """Creates the shell script that is passed to `sbatch`."""
        slurm_dir = self.slurm_io_dir
        repo_dir = path.dirname(path.dirname(path.dirname(__file__)))
        apptainer_script = shlex.quote(path.join(repo_dir, 'llacie/batch/sh/apptainer_slurm.sh'))

        if self.worker_cache is not None:
            cache_file = self.worker_cache.get_path("$SLURM_ARRAY_JOB_ID", "$SLURM_ARRAY_TASK_ID",
                False)
        else: cache_file = "-"

        return self.job_template.format(
            venv_activate = self.venv_activate, 
            run_worker = self.worker_cmd,
            ids_file = f"\"{slurm_dir}/job.$SLURM_ARRAY_JOB_ID.$SLURM_ARRAY_TASK_ID.ids\"",
            cache_file = cache_file,
            apptainer_script = apptainer_script,
            apptainer_image = shlex.quote(self.apptainer_image),
            apptainer_data_volume = shlex.quote(self.apptainer_data_volume)
        )
    

    def _parse_scontrol_job_state(self, stanza):
        lines = stanza.split("\n")
        if len(lines) >= 4:
            pairs = lines[3].strip().split()
            fields = {}
            for pair in pairs:
                if '=' in pair:
                    key, value = pair.split('=', 1)
                    fields[key] = value
            if "JobState" in fields: return fields["JobState"] == "RUNNING"
        return None


    def is_running(self):
        if self.array_job_id is None: return None
        cmd = f"scontrol show job {self.array_job_id}"
        proc = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.PIPE, shell=True, text=True)
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            if "slurm_load_jobs error: Invalid job id" in stderr: return False
        else:
            stanzas = stdout.strip().split("\n\n")
            statuses = [self._parse_scontrol_job_state(stanza) for stanza in stanzas]
            if not any(status is None for status in statuses):
                return any(status is True for status in statuses)
        echo_warn("Problem interrogating job status for Slurm job {self.array_job_id}")
        echo_warn(f"stdout:\n{stdout}\nstderr:{stderr}")
        return None
    

    def _sweep_worker_cache(self):
        count = self.worker_cache.sweep_into_upstream_db()
        self._done_counter += count
        if count > 0:
            plural = "s have" if self._done_counter != 1 else " has"
            echo_info(f"{self._done_counter} worker dataset{plural} now been saved "
                f"back into Postgres.")


    def run(self, all_ids):
        """
        Kicks off this job array in Slurm, and if a WorkerCache is needed, monitors the status of 
        the interim files and merges results back into Postgres until the job is completed.
        """
        id_chunks = chunker(all_ids, self.job_size)

        if len(id_chunks) == 0:
            return echo_err("There are no notes to process. Exiting.")
        if self.worker_cache is not None:
            self.worker_cache.cache_note_sections(id_chunks)

        self.array_job_id = self._run_background(id_chunks)

        if self.worker_cache is not None and self.array_job_id is not None:
            echo_info("Watching the worker cache directory for results from the workers")
            self.worker_cache.rename(self.array_job_id)
            
            time.sleep(self.WORKER_CACHE_POLL_INTERVAL)
            while self.is_running() is not False:
                self._sweep_worker_cache()
                time.sleep(self.WORKER_CACHE_POLL_INTERVAL)
            
            self._sweep_worker_cache()
            echo_info("The Slurm job has completed!")


import os
import re
import time
import subprocess as sub
import shlex

from os import path
from click import echo, secho
from textwrap import dedent

from ..utils import chunker, strtobool, echo_err, echo_info

def eprint(msg):
    echo(msg, err=True)
    secho("", nl=False, err=True, reset=True)

class SlurmJob(object):
    """
    Encapsulates an invocation of srun, managing its input, output, and process state, and
    polling for completion. We have to supervise srun like this if we want to start slurm/srun
    jobs in interactive mode (srun then expects to be able to read and write from pipes to a 
    parent process, and if they close, the job gets terminated.)
    """
    slurm_count = 0
    SLURM_PATH = "srun -v --pty"
    SLURM_SHELL = "/bin/bash"

    def __init__(self, slurm_opts, slurm_script, total_jobs, job_prefix="llama_", autostart=True):
        self.bin = SlurmJob.SLURM_PATH
        self.slurm_opts = slurm_opts
        self.slurm_script = slurm_script
        self.job_prefix = job_prefix

        SlurmJob.slurm_count += 1
        self.proc = None
        self.job_id = SlurmJob.slurm_count
        self.log_prefix = f"[job {self.job_id}/{total_jobs}] "
        self.slurm_job_id = None
        self.slurm_partition = None
        self.slurm_host = None

        self.stdout_lines = []
        self.stderr_lines = []
        self.stderr_log = []

        if autostart: self.start()

    def start(self):
        shell = SlurmJob.SLURM_SHELL
        self.cmd = f"{self.bin} {self.slurm_opts} -J {self.job_prefix}{self.job_id} {shell}"
        eprint(self.cmd)
        self.proc = sub.Popen(self.cmd, stdin = sub.PIPE, stdout = sub.PIPE, stderr = sub.PIPE, 
            shell = True)
        os.set_blocking(self.proc.stdout.fileno(), False)
        os.set_blocking(self.proc.stderr.fileno(), False)
        try: 
            outs, errs = self.proc.communicate(input = self.slurm_script.encode(), timeout = 1)
        except sub.TimeoutExpired as e:
            outs, errs = e.stdout, e.stderr
        self.stdout_lines = re.split(b'(?<=\n)', outs if outs is not None else b'')
        self.stderr_lines = re.split(b'(?<=\n)', errs if errs is not None else b'')
        self.process_lines()
        self.proc.stdin.close()

    def _eprint_submitted_msg(self):
        if self.slurm_partition is None or self.slurm_job_id is None: return
        eprint(f"{self.log_prefix}Submitted to slurm partition <{self.slurm_partition}> "
            f"job id {self.slurm_job_id}")

    def process_line(self, line, from_stderr=False):
        if from_stderr and (m := re.match(r'^srun: partition\s*:\s*(\w+)', line)):
            self.slurm_partition = m[1]
            self._eprint_submitted_msg()
        elif from_stderr and (m := re.match(r'^srun: job (\d+) queued and waiting', line)):
            self.slurm_job_id = int(m[1])
            self._eprint_submitted_msg()
        elif from_stderr and (m := re.match(r'^srun: jobid (\d+)', line)):
            self.slurm_job_id = int(m[1])
            self._eprint_submitted_msg()
        elif from_stderr and (m := re.match(r'^srun: launch/slurm: _task_start: Node (\w+)', line)):
            self.slurm_host = m[1]
            eprint(f"{self.log_prefix}Starting on {self.slurm_host}")
        elif not from_stderr and re.match(r'^srun: ', line) is None:
            eprint(f"{self.log_prefix}{line}")

    def process_lines(self, including_incomplete=False):
        for i, line in enumerate(self.stdout_lines):
            if not line.endswith(b"\n") and not including_incomplete: continue
            self.process_line(line.decode().strip(), False)
            self.stdout_lines.pop(i)
        for i, line in enumerate(self.stderr_lines):
            if not line.endswith(b"\n") and not including_incomplete: continue
            line = line.decode().strip()
            self.process_line(line, True)
            self.stderr_log.append(line)
            self.stderr_lines.pop(i)

    def flush_streams(self):
        while not self.proc.stdout.closed:
            line = self.proc.stdout.readline()
            if len(line) == 0: break
            if len(self.stdout_lines) > 0 and not self.stdout_lines[-1].endswith(b"\n"): 
                self.stdout_lines[-1] += line
            else: self.stdout_lines.append(line)
        while not self.proc.stderr.closed:
            line = self.proc.stderr.readline()
            if len(line) == 0: break
            if len(self.stderr_lines) > 0 and not self.stderr_lines[-1].endswith(b"\n"):
                self.stderr_lines[-1] += line
            else: self.stderr_lines.append(line)
        self.process_lines()
    
    def is_done(self):
        if self.proc.poll() is not None:
            self.flush_streams()
            if self.proc.returncode != 0:
                stderr_log = "\n".join(self.stderr_log)
                eprint(f"{self.log_prefix}Slurm salloc invocation failed.\n"
                    f"---\n{stderr_log}\n---")
                #TODO: figure out why it failed, and if we need to retry this job?
            else:
                eprint(f"{self.log_prefix}Done; success!")
            return True
        return False
    

class SlurmJobManager(object):
    """
    Manages a series of SlurmJob's over a set of IDs. These can be note IDs, episode IDs, or
    patient IDs, but in any of these cases, this Manager will kick off a series of SlurmJob's
    that run `worker_cmd` followed by sets of IDs as additional arguments, with the 
    concurrency and resource requests specified in either the constants within this class
    (defaults) or the configuration object passed in during instantiation.
    """
    SLURM_INTERACTIVE = False
    SLURM_SBATCH_PATH = "sbatch"
    SLURM_IO_DIR = os.path.join(os.environ["HOME"], "slurm")
    SLURM_IDS_PER_JOB = 100
    SLURM_CONCURRENT_JOBS = 1   # <-- Note that this only affects interactive job management
    SLURM_PARTITION = "long,bigmem,normal"
    SLURM_VENV = '.venv'
    SLURM_SBATCH_TIMEOUT = 60

    # Quick refresher on common options for Slurm's sbatch and srun:
    # -p = Slurm partition (aka queue)
    # -n = how many processes
    #      (note that the number of cores per process is set in `_create_slurm_job()` with -c)
    # -N = how many nodes (min-max or one number)
    # -t = time limit in HH:MM:SS
    # --mem = total memory limit, suffixed in K/M/G/T
    # -x = exclude certain nodes by name (e.g., if they are misbehaving)
    SLURM_OPTS = "-n 1 -N 1 -t 24:00:00 --mem=12G -x dn078"
    LLAMA_CPP_THREADS = 16
    SLURM_JOB_TEMPLATE = dedent("""\
        #!/bin/bash
        source ~/.bashrc
        conda activate llacie
        source {venv_activate}
        {run_worker} {ids}
        exit
        """)

    def __init__(self, worker_cmd, config=dict()):
        if not isinstance(worker_cmd, str):
            worker_cmd = " ".join([shlex.quote(str(piece)) for piece in worker_cmd])
        self.worker_cmd = worker_cmd
        repo_dir = path.dirname(path.dirname(path.dirname(__file__)))

        self.slurm_venv = config.get("SLURM_VENV", self.SLURM_VENV)
        self.venv_activate = path.join(repo_dir, f"{self.slurm_venv}/bin/activate")

        self.slurm_interactive = strtobool(config.get("SLURM_INTERACTIVE", self.SLURM_INTERACTIVE))  
        self.slurm_sbatch = config.get("SLURM_SBATCH_PATH", self.SLURM_SBATCH_PATH)
        self.slurm_io_dir = config.get("SLURM_IO_DIR", self.SLURM_IO_DIR)
        self.slurm_opts = config.get("SLURM_OPTS", self.SLURM_OPTS)
        self.slurm_partition = config.get("SLURM_PARTITION", self.SLURM_PARTITION)
        self.job_size = int(config.get("SLURM_IDS_PER_JOB", self.SLURM_IDS_PER_JOB))
        self.concurrent_jobs = int(config.get("SLURM_CONCURRENT_JOBS", self.SLURM_CONCURRENT_JOBS))
        self.job_template = config.get("SLURM_JOB_TEMPLATE", self.SLURM_JOB_TEMPLATE)
        self.threads = int(config.get("SLURM_THREADS_PER_JOB", 
            config.get("LLAMA_CPP_THREADS", self.LLAMA_CPP_THREADS)))

        self._slurm_jobs = []


    def _create_interactive_job(self, ids, total_jobs):
        ids = " ".join([shlex.quote(str(id_)) for id_ in ids])
        slurm_script = self.job_template.format(
            venv_activate = self.venv_activate, 
            run_worker = self.worker_cmd,
            ids = ids
        )
        slurm_opts = f"-c {str(self.threads)} -p {self.slurm_partition} {self.slurm_opts}"
        self._slurm_jobs.append(SlurmJob(slurm_opts, slurm_script, total_jobs, 'llacie_'))


    def _run_interactive(self, id_chunks):
        total_jobs = len(id_chunks)
        while True:
            # reap any jobs that have terminated
            for i, job in enumerate(self._slurm_jobs):
                job.flush_streams()
                if job.is_done(): self._slurm_jobs.pop(i)
            # pause to let things settle
            time.sleep(1.0)
            # sow any new jobs that need to be started
            while len(id_chunks) > 0 and len(self._slurm_jobs) < self.concurrent_jobs:
                ids = id_chunks.pop(0)
                self._create_interactive_job(ids, total_jobs)
            # if there are no jobs running, and no jobs left to start, we are done.
            if len(self._slurm_jobs) == 0: break

    
    def _create_sbatch_script(self):
        """Creates the shell script that is passed to `sbatch` for background slurm jobs."""
        slurm_dir = self.slurm_io_dir

        return self.job_template.format(
            venv_activate = self.venv_activate, 
            run_worker = self.worker_cmd,
            ids = f"$(cat \"{slurm_dir}/job.$SLURM_ARRAY_JOB_ID.$SLURM_ARRAY_TASK_ID.ids\")"
        )


    def _run_background(self, id_chunks):
        """Runs `worker_cmd` on each chunk of ids in `id_chunks` using a Slurm job array."""
        total_jobs = len(id_chunks)
        slurm_dir = self.slurm_io_dir
        os.makedirs(slurm_dir, exist_ok = True)

        slurm_opts = (f"{self.slurm_opts} -J llacie --array=1-{total_jobs} "
            + f"-o \"{slurm_dir}/job.%A.%a.out\" -e \"{slurm_dir}/job.%A.%a.err\" "
            + f"-c {str(self.threads)} -p {self.slurm_partition}")
        slurm_script = self._create_sbatch_script()

        proc = sub.Popen(f"{self.slurm_sbatch} {slurm_opts}", stdin = sub.PIPE, 
            stdout = sub.PIPE, stderr = sub.PIPE, shell = True)
        stdout, stderr = proc.communicate(input = slurm_script.encode(), 
            timeout = self.SLURM_SBATCH_TIMEOUT)
        match = re.match(r'^Submitted batch job (\d+)', stdout.decode().strip())

        if match is None: 
            echo_err(f"Error: unexpected sbatch output: {stdout.decode()}\n{stderr.decode()}")
            return None

        array_job_id = int(match[1])
        for i, ids in enumerate(id_chunks):
            with open(f"{slurm_dir}/job.{array_job_id}.{i + 1}.ids", 'w') as f:
                f.write(" ".join([str(j) for j in ids]))
        echo_info(f"Submitted {total_jobs} jobs under Slurm array job ID {array_job_id}")
        return array_job_id


    def run(self, all_ids):
        id_chunks = chunker(all_ids, self.job_size)
        if len(id_chunks) == 0:
            return echo_err("There are no notes to process. Exiting.")
        if self.slurm_interactive:
            self._run_interactive(id_chunks)
        else:
            self._run_background(id_chunks)


import os
import re
import json

from llama_cpp import Llama
from transformers import AutoTokenizer
from json.decoder import JSONDecodeError
from tqdm import tqdm

from ..utils import echo_err

class LlamaCppPython(object):
    """
    Runs LLMs using the `llama-cpp-python` package, which creates Python bindings for llama.cpp.
    Structured outputs are supported using constrained decoding and JSON schemas.
    See: https://github.com/abetlen/llama-cpp-python?tab=readme-ov-file#json-schema-mode
    """
    
    HOME_DIR = os.path.expanduser('~')
    ARRAY_OF_STRINGS_JSON_SCHEMA = """
    {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "array",
        "items": {
            "type": "string"
        }
    }
    """
    JSON_SCHEMA = ARRAY_OF_STRINGS_JSON_SCHEMA


    def __init__(self, config, model_path, system_prompt, user_prompt, json_schema=None, 
            tokenizer_model_path=None, **engine_args):
        """
        Instantiates an instance of the engine. For a comprehensive list of `engine_args` see
        the arguments for the Llama() constructor:
        https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.Llama
        """
        self.config = config
        self.model_path = model_path
        self.system_prompt = system_prompt
        self.user_prompt = user_prompt
        self.json_schema = json_schema if json_schema is not None else self.JSON_SCHEMA

        self.tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_model_path, 
            local_files_only = tokenizer_model_path.count('/') > 1
        )

        # If provided with the "filename" engine argument, the model_path is a HuggingFace repo ID
        if engine_args.get("filename"):
            # In this case, the model is loaded from HuggingFace
            self._model = Llama.from_pretrained(model_path, **engine_args)
        else:
            self._model = Llama(model_path, **engine_args)

    
    def create_prompt(self, **kwargs):
        messages = [
            {"role": "system", "content": self.system_prompt.format(**kwargs)},
            {"role": "user", "content": self.user_prompt.format(**kwargs)}
        ]
        return messages


    def autotrim_prompt(self, prompt_max_tokens, trim_key="input", **kwargs):
        """Creates a prompt and ensures it is within the token limit.

        `kwargs` are substituted into the chat template created by `system_prompt` and
        by `user_prompt`, using Python's native str.format().

        `prompt_max_tokens` is the token limit, usually $context_window - 4.

        If the generated prompt is longer than `prompt_max_tokens` after tokenization, the item 
        in the dict() `kwargs` at key `trim_key` is trimmed by about one sentence at a time until 
        the threshold is reached. By default, `kwargs["input"]` is trimmed.
        
        Returns the full prompt with the trimmed inputs inserted, or None if the process fails."""
        trimmed_text = kwargs.get(trim_key, "")
        if self.tokenizer is None:
            echo_err("A tokenizer for this model could not be loaded.")
            return None
        while True:
            full_prompt = self.create_prompt(**kwargs)
            tokenized_prompt = self.tokenizer.apply_chat_template(
                full_prompt,
                tokenize = True, 
                add_generation_prompt = True
            )
            if len(tokenized_prompt) <= prompt_max_tokens: break
            # If it's too long, cleave off the last sentence-like piece, and try again
            sentences_and_periods = re.split(r'([.]\s+|[.]$)', trimmed_text)
            if len(sentences_and_periods) < 4:
                return None
            else: 
                trimmed_text = ''.join(sentences_and_periods[:-4]) + '.'
                kwargs[trim_key] = trimmed_text
        return full_prompt
    

    def __call__(self, prompts, **sampling_params):
        response_format = {
            "type": "json_object",
            "schema": json.loads(self.json_schema),
        }

        outputs = []
        for prompt in tqdm(prompts):
            try:
                full_output = self._model.create_chat_completion(
                    prompt,
                    response_format=response_format,
                    **sampling_params
                )
                output = json.loads(full_output["choices"][0]["message"]["content"])
                outputs.append(output)
            except (IndexError, KeyError):
                echo_err(f"Invalid response from create_chat_completion: {full_output}")
                outputs.append(None)
            except JSONDecodeError as e:
                echo_err(f"Invalid JSON. '{e.msg}' at char {e.pos}:\n{e.doc}")
                outputs.append(None)

        return outputs


from os import path
from glob import glob
from fnmatch import fnmatch
from importlib import import_module
from inspect import getmembers

from .abstract import AbstractStrategy
from ..tasks.abstract import TaskOutputType as out_t, AbstractTask

class StrategyFinder:
    _index = None
    GLOB = '*/*/*.py'
    BASE_PATH = path.dirname(__file__)

    def _create_index(self):
        """Creates an index of the modules/classes within this package
        for use by the `find_strategies` method below."""
        self._index = []
        already_imported = set()

        for py_file in glob(path.join(self.BASE_PATH, self.GLOB)):
            if path.basename(py_file).startswith('__'): continue

            py_file = py_file[len(self.BASE_PATH):]
            module_name = path.splitext(py_file)[0].replace(path.sep, '.')
            mod = import_module(module_name, package=__name__)

            for name, obj in getmembers(mod):
                if name.startswith('_'): continue
                if obj is AbstractStrategy or name.startswith('Abstract'): continue
                if obj in already_imported: continue
                if not isinstance(obj, type) or not issubclass(obj, AbstractStrategy):
                    continue

                already_imported.add(obj)
                self._index.append({
                    "output_type": obj.task.output_type,
                    "task": obj.task, 
                    "task_name": obj.task.name,
                    "strategy_name": obj.name,
                    "class_name": name,
                    "strategy": obj
                })

    @property
    def index(self):
        """Lazily create the indexes, once, upon first access"""
        if self._index is not None: return self._index
        self._create_index()
        return self._index
    
    def name_glob_match(self, full_name, name_glob):
        if name_glob.count(".") < 2: name_glob = f"*.{name_glob}"
        return fnmatch(full_name, name_glob)
    
    def find_strategies(self, output_type=None, task=None, name_glob=None):
        strategies = []

        for row in self.index:
            if name_glob is not None: 
                if not self.name_glob_match(row["strategy_name"], name_glob): continue

            if output_type is not None:
                if isinstance(output_type, str):
                    if row["output_type"].value != output_type: continue
                else:
                    if not isinstance(output_type, out_t):
                        raise RuntimeError("output_type must be `str` or `TaskOutputType`")
                    if output_type is not row["output_type"]: continue

            if task is not None:
                if isinstance(task, str):
                    if row["task_name"] != task: continue
                else:
                    if not issubclass(task, AbstractTask):
                        raise RuntimeError("task must be `str` or subclass of `AbstractTask`")
                    if row["task"] is not task: continue

            strategies.append(
                (row["output_type"], row["task_name"], row["strategy_name"], row["strategy"]))

        strategies = [row[3] for row in sorted(strategies, key = lambda row: row[0:3])]
        return strategies


# We create one global instance of the above class
_finder = StrategyFinder()

# Export this prebound method
find_strategies = _finder.find_strategies


from .llama3_8b import PresentingSxFeatureLlama3Instruct8BStrategy
from ...section.hpi_short.regex import ShortHPISectionRegexStrategy
from ....tasks.feature import PresentingSymptomsFeatureTask

class PresentingSxFeatureLlama31Instruct8BStrategy(PresentingSxFeatureLlama3Instruct8BStrategy):
    """\
    Attempts to extract the `presenting_sx` feature using Llama 3.1 Instruct 8B model,
    an 8B parameter model in the Llama family released by Meta in July 2024. 

    This uses the original weights casted to torch.float16, because the Tesla V100s that we can 
    run this on in ERISXdl do not support bfloat16. 
    https://huggingface.co/meta-llama/Meta-Llama-3.1-8B-Instruct

    Also uses the `hpi_short` sections created by the section.hpi_short.regex strategy.
    """
    task = PresentingSymptomsFeatureTask
    name = "feature.presenting_sx.llama31_8b"
    version = "0.0.1"
    prereq_tasks = []

    SECTION_STRATEGY_CLASS = ShortHPISectionRegexStrategy

    MODEL_ID = 'meta-llama/Meta-Llama-3.1-8B-Instruct'

    # Model context length. Note that Vllm will not generate past the end of the context window.
    # Llama 3.1 supports up to 131,072 (128k) but this is also constrained by GPU memory
    LLM_MAX_MODEL_LENGTH = 4096 

    NICE_MODEL_NAME = "Llama3.1 8B Instruct"


from tqdm import tqdm

from ...abstract import AbstractStrategy
from ....tasks.episode_label import PresentingSymptomsEpisodeLabelV2Task
from ....utils import chunker, echo_info

SECS_IN_24H = 60 * 60 * 24

class AbstractPresentingSymptomsEpisodeLabelV2Strategy(AbstractStrategy):
    """\
    An abstract strategy for converting presenting_sx note features into episode labels using 
    vocab v2. This vocab is UPDATED since the K08 submission after manual examination of the
    misclassifications.
    """
    task = PresentingSymptomsEpisodeLabelV2Task
    prereq_tasks = []

    BATCH_SIZE = 1000


    def run(self, all_episode_ids, batch_size = None):
        fail_count = 0
        vocab = self.task.vocab
        batch_size = batch_size if batch_size is not None else self.BATCH_SIZE

        echo_info(f"In batches of {batch_size}")
        needs_labels_pb = tqdm(chunker(all_episode_ids, batch_size))

        for ep_ids in needs_labels_pb:
            needs_labels_pb.set_description(
                f"Creating labels ({fail_count} failures/{len(all_episode_ids)} total)")
            
            df = self.db.get_earliest_notes_with_feature(ep_ids, self.feat_strat, SECS_IN_24H)
            # Some episodes may not have any notes with the required feature
            fail_count += len(ep_ids) - len(df)

            for _, row in df.iterrows():
                labels_dict = vocab.find_terms_in_feature(row.feature_value)

                if len(labels_dict) == 0: 
                    fail_count += 1
                else:
                    self.db.replace_episode_labels(self, row, labels_dict)

        echo_info(f"Finished! ({fail_count} failures/{len(all_episode_ids)} total)")



import os
import re
import time
import subprocess as sub
import shlex

from os import path
from click import echo, secho
from textwrap import dedent

from ..utils import chunker

def eprint(msg):
    echo(msg, err=True)
    secho("", nl=False, err=True, reset=True)

class BsubJob(object):
    """
    Encapsulates an invocation of bsub, managing its input, output, and process state, and
    polling for completion. We have to supervise bsub like this if we want to start bsub
    jobs in interactive mode (bsub expects to be able to read and write from pipes to a 
    parent process, and if they close, the job gets terminated.)
    """
    bsub_count = 0
    BSUB_PATH = "bsub"

    def __init__(self, bsub_opts, bsub_script, total_jobs, job_prefix="llama_", autostart=True):
        self.bin = BsubJob.BSUB_PATH
        self.bsub_opts = bsub_opts
        self.bsub_script = bsub_script
        self.job_prefix = job_prefix

        BsubJob.bsub_count += 1
        self.proc = None
        self.job_id = BsubJob.bsub_count
        self.log_prefix = f"[job {self.job_id}/{total_jobs}] "
        self.lsf_job_id = None
        self.lsf_queue = None
        self.lsf_host = None

        if autostart: self.start()

    def start(self):
        self.cmd = f"{self.bin} {self.bsub_opts} -J {self.job_prefix}{self.job_id}"
        self.proc = sub.Popen(self.cmd, stdin = sub.PIPE, stdout = sub.PIPE, stderr = sub.PIPE, 
            shell = True)
        os.set_blocking(self.proc.stdout.fileno(), False)
        os.set_blocking(self.proc.stderr.fileno(), False)
        try: 
            outs, errs = self.proc.communicate(input = self.bsub_script.encode(), timeout = 1)
        except sub.TimeoutExpired as e:
            outs, errs = e.stdout, e.stderr
        self.stdout_lines = re.split(b'(?<=\n)', outs if outs is not None else b'')
        self.stderr_lines = re.split(b'(?<=\n)', errs if errs is not None else b'')
        self.process_lines()
        self.proc.stdin.close()
    
    def process_line(self, line, from_stderr=False):
        if (m := re.match(r'Job <(\d+)> is submitted to queue <(\w+)>', line)) and not from_stderr:
            self.lsf_job_id = int(m[1])
            self.lsf_queue = m[2]
            eprint(f"{self.log_prefix}Submitted to <{self.lsf_queue}>, id {self.lsf_job_id}")
        elif (m := re.match(r'<<Starting on (\w+)>>', line)) and from_stderr:
            self.lsf_host = m[1]
            eprint(f"{self.log_prefix}Starting on {self.lsf_host}")
        elif not from_stderr:
            eprint(f"{self.log_prefix}{line}")

    def process_lines(self, including_incomplete=False):
        for i, line in enumerate(self.stdout_lines):
            if not line.endswith(b"\n") and not including_incomplete: continue
            self.process_line(line.decode().strip(), False)
            self.stdout_lines.pop(i)
        for i, line in enumerate(self.stderr_lines):
            if not line.endswith(b"\n") and not including_incomplete: continue
            self.process_line(line.decode().strip(), True)
            self.stderr_lines.pop(i)

    def flush_streams(self):
        while True:
            line = self.proc.stdout.readline()
            if len(line) == 0: break
            if len(self.stdout_lines) > 0 and not self.stdout_lines[-1].endswith(b"\n"): 
                self.stdout_lines[-1] += line
            else: self.stdout_lines.append(line)
        while True:
            line = self.proc.stderr.readline()
            if len(line) == 0: break
            if len(self.stderr_lines) > 0 and not self.stderr_lines[-1].endswith(b"\n"):
                self.stderr_lines[-1] += line
            else: self.stderr_lines.append(line)
        self.process_lines()
    
    def is_done(self):
        if self.proc.poll() is not None:
            self.flush_streams()
            if self.proc.returncode != 0:
                eprint(f"{self.log_prefix}Bsub invocation failed.\n---\n{self.bsub_script}\n---")
                #TODO: figure out why it failed, and if we need to retry this job?
            else:
                eprint(f"{self.log_prefix}Done; success!")
            return True
        return False
    

class BsubJobManager(object):
    """
    Manages a series of BsubJob's over a set of IDs. These can be note IDs, episode IDs, or
    patient IDs, but in any of these cases, this Manager will kick off a series of BsubJob's
    that run `worker_cmd` followed by sets of IDs as additional arguments, with the 
    concurrency and resource requests specified in either the constants within this class
    (defaults) or the configuration object passed in during instantiation.
    """
    BSUB_IDS_PER_JOB = 50
    BSUB_CONCURRENT_JOBS = 10
    BSUB_PATH = BsubJob.BSUB_PATH
    BSUB_OPTS = "-Is -q interactive"
    LLAMA_CPP_THREADS = 16
    BSUB_LSF_JOB = dedent("""\
        #!/bin/bash
        #BSUB -W 24:00
        #BSUB -R 'rusage[mem=12000]'
        #BSUB -R 'span[hosts=1]'
        #BSUB -L /bin/bash

        source ~/.bashrc
        conda activate llacie
        source {venv_activate}
        {run_worker} {ids}
        """)

    def __init__(self, worker_cmd, config=dict()):
        if not isinstance(worker_cmd, str):
            worker_cmd = " ".join([shlex.quote(str(piece)) for piece in worker_cmd])
        self.worker_cmd = worker_cmd
        repo_dir = path.dirname(path.dirname(__file__))
        self.venv_activate = path.join(repo_dir, ".venv/bin/activate")
        self.worker_cmd = worker_cmd
        
        self.bsub_path = config.get("BSUB_PATH", self.BSUB_PATH)
        self.bsub_opts = config.get("BSUB_OPTS", self.BSUB_OPTS)
        self.job_size = int(config.get("BSUB_IDS_PER_JOB", self.BSUB_IDS_PER_JOB))
        self.concurrent_jobs = int(config.get("BSUB_CONCURRENT_JOBS", self.BSUB_CONCURRENT_JOBS))
        self.job_template = config.get("BSUB_LSF_JOB", self.BSUB_LSF_JOB)
        self.threads = int(config.get("BSUB_THREADS_PER_JOB", 
            config.get("LLAMA_CPP_THREADS", self.LLAMA_CPP_THREADS)))

        self._bsub_jobs = []


    def _create_bsub_job(self, ids, total_jobs):
        ids = " ".join([shlex.quote(str(id_)) for id_ in ids])
        bsub_script = self.job_template.format(
            venv_activate = self.venv_activate, 
            run_worker = self.worker_cmd,
            ids = ids
        )
        bsub_opts = f"-n {str(self.threads)} {self.bsub_opts}"
        self._bsub_jobs.append(BsubJob(bsub_opts, bsub_script, total_jobs, 'llama_'))


    def run(self, all_ids):
        id_chunks = chunker(all_ids, self.job_size)
        total_jobs = len(id_chunks)
        while True:
            # reap any jobs that have terminated
            for i, job in enumerate(self._bsub_jobs):
                job.flush_streams()
                if job.is_done(): self._bsub_jobs.pop(i)
            # pause to let things settle
            time.sleep(1.0)
            # sow any new jobs that need to be started
            while len(id_chunks) > 0 and len(self._bsub_jobs) < self.concurrent_jobs:
                ids = id_chunks.pop(0)
                self._create_bsub_job(ids, total_jobs)
            # if there are no jobs running, and no jobs left to start, we are done.
            if len(self._bsub_jobs) == 0: break


import subprocess as sub
import shlex
import time

from os import path
from textwrap import dedent

from .slurm import SlurmJobManager
from ..utils import chunker, echo_warn, echo_info, echo_err
from ..cache.abstract import AbstractFeatureWorkerCache

class KubeSlurmJobManager(SlurmJobManager):
    """
    Uses slurm in the specific manner expected by ERISXdl, in which a wrapper script is needed for 
    each job to reserve a GPU, deploy a container from Podman onto the node, and then run all 
    commands within the container.

    This is documented at https://rc.partners.org/kb/computational-resources/erisxdl?article=3719

    The wrapper script is at /data/erisxdl/kube-slurm/wrappers/kube-slurm-custom-image-job.sh
    """

    # The smallest walltime limit for all paid queues is 2 hours.
    # At about ~1 sec per ID, a 2000-ID job takes ~40min
    SLURM_IDS_PER_JOB = 2000
    SLURM_PARTITION = "Short,Medium,Long"
    SLURM_VENV = '.venv-xdl'  # <-- relative to the root of the repo
    SLURM_THREADS_PER_JOB = 10
    # Adds a random wait 
    SLURM_MAX_RANDOM_WAIT = 60

    # Configuration that is passed into the wrapper script
    KUBE_IMAGE = 'erisxdl.partners.org/bwh-tide_eris1-g/cuda:vllm-llacie' # Points to a podman image
    KUBE_DATA_VOLUME = '/data/tide'
    KUBE_INIT_TIMEOUT = 600
    KUBE_WRAPPER_SCRIPT = '/data/erisxdl/kube-slurm/wrappers/kube-slurm-custom-image-job.sh'

    # Currently, ERISXdl GPU nodes don't have any outbound networking at all, not even allowing
    # connections to the Postgres database. As a workaround, we use a filesystem-based cache as an 
    # intermediary. This WorkerCache functionality is implemented in the llacie.cache.* modules
    KUBE_WORKERS_HAVE_NETWORKING = False
    WORKER_CACHE_POLL_INTERVAL = 15

    # Quick refresher on common options for Slurm's sbatch and srun:
    # -p = Slurm partition (aka queue)
    # -n, --ntasks = how many processes
    #      (note that the number of cores per process is set in `_create_slurm_job()` with -c)
    # -N = how many nodes (min-max or one number)
    # -t = time limit in HH:MM:SS
    # --mem = total memory limit, suffixed in K/M/G/T
    # -x = exclude certain nodes by name (e.g., if they are misbehaving)
    SLURM_OPTS = "-n 1 -t 2:00:00 --gpus=1 --mem=16G"

    SLURM_JOB_TEMPLATE = dedent("""\
        #!/bin/bash
        ## Set the docker container image to be used in the job runtime.
        export KUBE_IMAGE={kube_image}

        ## Set the script to be run within the specified container - this MUST be a separate script
        export KUBE_SCRIPT={kube_script}
        ## Other vars that the script can use
        ids_file={ids_file}
        cache_file={cache_file}
        venv_activate={venv_activate}
        export KUBE_SCRIPT_VARS="$venv_activate $cache_file $ids_file {run_worker}"

        # Define group briefcase (this will provide the GID for the user at runtime)
        export KUBE_DATA_VOLUME={kube_data_volume}

        # Users can also set the following variable to change the timeout in seconds.
        # It’s 600 by default, but might be useful to change for testing.
        export KUBE_INIT_TIMEOUT={kube_init_timeout}
                                
        # Introduced a random sleep here due to issues mounting /data/tide within the container
        # See ticket #INC4206564 with Richard Kenny. Seems very like a very fragile solution...
        MAX_RANDOM_WAIT={max_random_wait}
        sleep $((RANDOM % MAX_RANDOM_WAIT))

        ## Required wrapper script. This must be included at the end of the job submission script.
        ## This wrapper script mounts /data, and your /PHShome directory into the container 
        ## It is provided and maintained by the ERISXdl admins.
        srun {kube_wrapper_script}
        """)


    def __init__(self, worker_cmd, config=dict(), worker_cache=None):
        super().__init__(worker_cmd, config)
        self.threads = int(config.get("SLURM_THREADS_PER_JOB", self.SLURM_THREADS_PER_JOB))
        self._max_rand_wait = int(config.get("SLURM_MAX_RANDOM_WAIT", self.SLURM_MAX_RANDOM_WAIT))

        self.kube_image = config.get("KUBE_IMAGE", self.KUBE_IMAGE)
        self.kube_data_volume = config.get("KUBE_DATA_VOLUME", self.KUBE_DATA_VOLUME)
        self.kube_init_timeout = int(config.get("KUBE_INIT_TIMEOUT", self.KUBE_INIT_TIMEOUT))
        self.kube_wrapper_script = config.get("KUBE_WRAPPER_SCRIPT", self.KUBE_WRAPPER_SCRIPT)

        # This JobManager does not ever support interactive jobs
        self.slurm_interactive = False
        self.array_job_id = None
        self._done_counter = 0

        # The behavior of this JobManager changes if the workers don't have any network capability
        # In this case, we use a filesystem-based cache
        self.workers_have_networking = config.get("KUBE_WORKERS_HAVE_NETWORKING", 
            self.KUBE_WORKERS_HAVE_NETWORKING)
        self.worker_cache = worker_cache
        if not self.workers_have_networking:
            if not isinstance(self.worker_cache, AbstractFeatureWorkerCache):
                raise RuntimeError("When providing a worker cache to a KubeSlurmJobManager, it "
                    "must inherit from AbstractFeatureWorkerCache")
            if self.worker_cache is None:
                raise RuntimeError("KubeSlurmJobManager requires a worker cache if the workers do "
                    "not have networking enabled!")


    # We override the creation of the sbatch script since it is more complicated with the 
    # introduction of the wrapper script.
    def _create_sbatch_script(self):
        """Creates the shell script that is passed to `sbatch`."""
        slurm_dir = self.slurm_io_dir
        repo_dir = path.dirname(path.dirname(path.dirname(__file__)))

        if self.worker_cache is not None:
            cache_file = self.worker_cache.get_path("$SLURM_ARRAY_JOB_ID", "$SLURM_ARRAY_TASK_ID",
                False)
        else: cache_file = "-"

        return self.job_template.format(
            max_random_wait = self._max_rand_wait,
            venv_activate = self.venv_activate, 
            run_worker = self.worker_cmd,
            ids_file = f"\"{slurm_dir}/job.$SLURM_ARRAY_JOB_ID.$SLURM_ARRAY_TASK_ID.ids\"",
            cache_file = cache_file,
            kube_script = shlex.quote(path.join(repo_dir, 'llacie/batch/sh/kube_slurm.sh')),
            kube_image = shlex.quote(self.kube_image),
            kube_data_volume = shlex.quote(self.kube_data_volume),
            kube_init_timeout = self.kube_init_timeout,
            kube_wrapper_script = shlex.quote(self.kube_wrapper_script)
        )
    

    def _parse_scontrol_job_state(self, stanza):
        lines = stanza.split("\n")
        if len(lines) >= 4:
            pairs = lines[3].strip().split()
            fields = {}
            for pair in pairs:
                if '=' in pair:
                    key, value = pair.split('=', 1)
                    fields[key] = value
            if "JobState" in fields: return fields["JobState"] == "RUNNING"
        return None


    def is_running(self):
        if self.array_job_id is None: return None
        cmd = f"scontrol show job {self.array_job_id}"
        proc = sub.Popen(cmd, stdout=sub.PIPE, stderr=sub.PIPE, shell=True, text=True)
        stdout, stderr = proc.communicate()
        if proc.returncode != 0:
            if "slurm_load_jobs error: Invalid job id" in stderr: return False
        else:
            stanzas = stdout.strip().split("\n\n")
            statuses = [self._parse_scontrol_job_state(stanza) for stanza in stanzas]
            if not any(status is None for status in statuses):
                return any(status is True for status in statuses)
        echo_warn("Problem interrogating job status for Slurm job {self.array_job_id}")
        echo_warn(f"stdout:\n{stdout}\nstderr:{stderr}")
        return None
    

    def _sweep_worker_cache(self):
        count = self.worker_cache.sweep_into_upstream_db()
        self._done_counter += count
        if count > 0:
            plural = "s have" if self._done_counter != 1 else " has"
            echo_info(f"{self._done_counter} worker dataset{plural} now been saved "
                f"back into Postgres.")


    def run(self, all_ids):
        """
        Kicks off this job array in Slurm, and if a WorkerCache is needed, monitors the status of 
        the interim files and merges results back into Postgres until the job is completed.
        """
        id_chunks = chunker(all_ids, self.job_size)

        if len(id_chunks) == 0:
            return echo_err("There are no notes to process. Exiting.")
        if self.worker_cache is not None:
            self.worker_cache.cache_note_sections(id_chunks)

        self.array_job_id = self._run_background(id_chunks)

        if self.worker_cache is not None and self.array_job_id is not None:
            echo_info("Watching the worker cache directory for results from the workers")
            self.worker_cache.rename(self.array_job_id)
            
            time.sleep(self.WORKER_CACHE_POLL_INTERVAL)
            while self.is_running() is not False:
                self._sweep_worker_cache()
                time.sleep(self.WORKER_CACHE_POLL_INTERVAL)
            
            self._sweep_worker_cache()
            echo_info("The Slurm job has completed!")


import sys
import json

from os import path
from textwrap import dedent
from socket import gethostname

from .abstract import AbstractStrategy
from ..batch.apptainer_slurm import ApptainerSlurmJobManager
from ..cache.sqlite import SqliteFeatureWorkerCache
from ..inference.vllm import Vllm
from ..utils import echo_warn, echo_info, strtobool

class AbstractVllmStrategy(AbstractStrategy):
    """
    Abstract class for running vllm with constant parameters and a prespecified prompt template, 
    across all of the specified notes, substituting one section for each note into the prompt 
    template, to extract a feature.

    The output is currently constrained to be a JSON array of strings. This could be overridden
    by inheriting classes, by redefining `self.JSON_SCHEMA`.
    
    Slurm MUST be used to schedule the jobs, given that ERISXdl requires use of both slurm and 
    Apptainer containers to access the GPUs on the GPU nodes.

    This class is intended to be subclassed and configured *primarily* by overriding the class 
    constants, and beyond that, its methods might be overridden as well.
    """
    # can be, e.g., ShortHPISectionRegexStrategy from .section.hpi_short.regex
    SECTION_STRATEGY_CLASS = None             

    # These are example default options based on the Llama 3 8B strategy for presenting symptoms
    MODEL_ID = 'meta-llama/Meta-Llama-3-8B-Instruct'
    SLURM_ENABLED = True
    SLURM_THREADS_PER_JOB = 10
    ARRAY_OF_STRINGS_JSON_SCHEMA = """
    {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "array",
        "items": {
            "type": "string"
        }
    }
    """
    JSON_SCHEMA = ARRAY_OF_STRINGS_JSON_SCHEMA

    # Need to set LLM(..., dtype='half') because Nvidia V100s do not natively support bfloat16
    # This is expected to trigger a warning: "Casting torch.bfloat16 to torch.float16"
    VLLM_DTYPE = 'half'

    LLM_TOKENS_TO_GENERATE = 256
    # Arguments to SamplingParams in `vllm`
    # See https://docs.vllm.ai/en/latest/dev/sampling_params.html
    SAMPLING_PARAMS = {
        "temperature": 0.1,
        "top_p": 0.8,
        "repetition_penalty": 1.05
    }

    NICE_MODEL_NAME = "Llama3 8B Instruct (vllm)"

    # Model context length. Note that Vllm will not generate past the end of the context window.
    LLM_MAX_MODEL_LENGTH = 2048  

    # Prompts
    LLM_SYSTEM_PROMPT = dedent("""\
        You are a clinical researcher that reads medical charts and answers questions about them. 
        Use only the information in the text provided to answer the question.
        If a patient denies something, do not include it in your answer.
        After you provide an answer, you immediately stop talking.""")
    LLM_USER_PROMPT = dedent("""\
        Read the following patient history and list the patient's presenting symptoms.
        Include only symptoms present now or reported for the days to weeks leading up to admission.
        Ignore any symptoms from past medical history or prior hospital admissions.
        Give your answer as a JSON array containing up to ten strings.
        Each string contains between one and three words.
        
        {input}""")


    def __init__(self, db, config, **options):
        super().__init__(db, config, **options)
        # Ensure this config option exists because ApptainerSlurmJobManager's parent class needs it
        self.config['SLURM_THREADS_PER_JOB'] = config.get("SLURM_THREADS_PER_JOB",
            self.SLURM_THREADS_PER_JOB)
        
        if not isinstance(self.SECTION_STRATEGY_CLASS, type):
            raise NotImplementedError("You must specify a SECTION_STRATEGY_CLASS")

        self.sec_strat = self.SECTION_STRATEGY_CLASS(db, config, **options)
        self.sec_task_name = self.sec_strat.task.name

        self.local_models_path = config.get("LOCAL_MODELS_PATH", None)
        self.model_path = self.MODEL_ID
        if self.local_models_path: self.model_path = f"{self.local_models_path}/{self.model_path}"

        self._sampling_params = {
            **self.SAMPLING_PARAMS, 
            "max_tokens": self.LLM_TOKENS_TO_GENERATE
        }
        self._dtype = config.get("VLLM_DTYPE", self.VLLM_DTYPE)

        self.slurm_enabled = strtobool(config.get("SLURM_ENABLED", self.SLURM_ENABLED))
        self.workers_have_networking = config.get("APPTAINER_WORKERS_HAVE_NETWORKING", 
            ApptainerSlurmJobManager.APPTAINER_WORKERS_HAVE_NETWORKING)

        if self.slurm_enabled and not self.workers_have_networking:
            self.worker_cache = SqliteFeatureWorkerCache(db, config, self)


    def _run_llm(self, note_ids):
        if self.within_worker and not self.workers_have_networking: db = self.worker_cache
        else: db = self.db

        llm = Vllm(self.config, self.model_path, self.LLM_SYSTEM_PROMPT, self.LLM_USER_PROMPT,
            self.JSON_SCHEMA, dtype=self._dtype, max_model_len=self.LLM_MAX_MODEL_LENGTH)
        max_prompt_length = self.LLM_MAX_MODEL_LENGTH - self.LLM_TOKENS_TO_GENERATE - 4

        self.start_timer()

        prompts = []
        section_ids = []
        for note_id in note_ids:
            # Get the note section to work on
            row = db.get_note_section(note_id, self.sec_strat)
            if row is None: 
                raise RuntimeError(f"No {self.sec_task_name} section found for note id {note_id}")
            section_value = row.section_value

            # Ensure the prompt will fit in the context window, with 4 tokens of wiggle room
            full_prompt = llm.autotrim_prompt(max_prompt_length, input=section_value)
            if full_prompt is None:
                echo_warn(f"Autotrimming failed for note id {note_id}")
                # Couldn't autotrim; likely a malformed section, but try using the whole thing
                full_prompt = llm.create_prompt(input=section_value)
            
            section_ids.append(row.id)
            prompts.append(full_prompt)
            
        # Run all the prompts through vllm
        # A key difference between this and llama.cpp is that we pipeline all prompts thru at once!
        # Also note that outputs are already deserialized from JSON into Python lists or dicts
        echo_info(f"Running {len(note_ids)} notes thru the {self.NICE_MODEL_NAME} model")
        outputs = llm(prompts, **self._sampling_params)
        runtime = self.stop_timer() / len(outputs)

        for i, output in enumerate(outputs):
            note_id = note_ids[i]
            section_id = section_ids[i]
            if output is not None and isinstance(output, list): 
                feature = "\n".join(output)
                output_raw = json.dumps(output)
                db.upsert_note_feature(note_id, self, output_raw, feature, section_id, runtime)
            else:
                echo_warn(f"No LLM output generated for note id {note_id}")


    def run(self, all_note_ids, max_note_ids=None, batch_size=None, dry_run=False):
        if self.within_worker:
            # We are within a worker process; run the LLM on the given chunk of note IDs
            if dry_run: return echo_info("Info: Dry run, exiting without running any LLMs")
            note_ids = all_note_ids
            echo_info(f"Worker started on {gethostname()}") 

        else:
            # There is no point running this on notes without the required section
            note_ids = self.db.filter_to_notes_with_section(all_note_ids, self.sec_strat)
            if (num_filtered := len(all_note_ids) - len(note_ids)) > 0:
                echo_warn(f"Skipping {num_filtered} notes without an {self.sec_task_name} section")
            if max_note_ids is not None and (len(note_ids) > max_note_ids):
                echo_warn(f"Limiting to the first {max_note_ids} notes")
                note_ids = note_ids[:max_note_ids]
            if dry_run: return echo_info("Info: Dry run, exiting without running any LLMs")
            
            if self.slurm_enabled:
                # Create a SlurmJobManager and submit jobs that run worker processes
                worker_cmd = sys.argv.copy()
                # We want a relative path to the `llacie` executable, because the worker processes
                #   may be running in a different virtualenv than this one
                worker_cmd[0] = path.basename(worker_cmd[0])
                # To start a worker, you invoke the same command with `--worker` appended
                worker_cmd.append("--worker")
                job_manager = ApptainerSlurmJobManager(worker_cmd, self.config, self.worker_cache)
                return job_manager.run(note_ids)
        
        # Either slurm is disabled or we are inside one the worker processes
        # In this case, we should start running vLLM on the note IDs
        self._run_llm(note_ids)

        # For worker processes without networking, we need to mark the results on disk as complete
        if self.within_worker and not self.workers_have_networking: self.worker_cache.finalize()


import re

# A version of this was incorporated into ShortHPISectionRegexStrategy
# It uses the circumstance that EDW replaced newlines with two spaces to
# reformat note text into a more readable form, by hardwrapping it

def fix_note_whitespace(text):
    text = re.sub("  ", "\n", text)
    text = re.sub("\n[ ?]+", "\n", text)
    text = re.sub("\n\n+", "\n\n", text)
    return text.strip()

##################################################################
# Functions for cleaning up list-like text spit out by LLMs
##################################################################

def split_listlike_text(raw_text):
    numeric_list = re.match(r'[1]([).]) +.+?([;,]?) [2]\1( +.+?(\2 |\2? and )\d+\1)*', raw_text)
    bulleted_list = re.match(r'([·•*-] ?)[^\n]+(\n\n?)\1\S', raw_text)

    if raw_text.startswith('\\begin{itemize}'):
        # LLaMA sometimes composes in LaTeX-style lists
        without_prefix = re.sub(r'^\\begin\{itemize\}\s+\\item +', '', raw_text)
        middle = re.sub(r'\s+\\end\{itemize\}[\s\S]*', '', without_prefix)
        values = re.split(r'\s+\\item +', middle)
    elif numeric_list is not None:
        without_suffix = raw_text
        # depending on the style of list, we look for the end of the sentence or paragraph
        end_regex = r'([.]\s+|\n\n)' if numeric_list[1] == ')' else r'[.]?\n\n'
        end_regex_match = re.search(end_regex, raw_text[numeric_list.span()[1]:])
        if end_regex_match is not None:
            without_suffix = raw_text[:numeric_list.span()[1] + end_regex_match.span()[0]]
        middle = re.sub(r'^\d[' + numeric_list[1] + r']', '', without_suffix)
        if numeric_list[2] != '':
            values = re.split(r'(?:' + numeric_list[2] + r'|' + numeric_list[2] + r'? and)? \d+[' + 
                    numeric_list[1] + r']', middle)
        else:
            values = re.split(r'(?: and)? \d+[' + numeric_list[1] + r']', middle)
    elif bulleted_list is not None:
        bullet_regex = r'[' + bulleted_list[1][0] + r']' + bulleted_list[1][1:]
        without_prefix = re.sub(r'^' + bullet_regex, '', raw_text)
        middle = re.sub(r'\n\n[^' + bulleted_list[1][0] + r'][\s\S]*', '', without_prefix)
        values = re.split(bulleted_list[2] + bullet_regex, middle)
    else:
        # Try treating it like a list within a sentence (fragment)
        # Also, excise any superfluous first bullets, numbers, etc
        first_sentence = re.sub(r'^(1[.)]|[·•*-])\s+', '', raw_text)
        first_sentence = re.sub(r'([.]\s+|[.]?\n\n)[\s\S]*', '', first_sentence)
        # If it's not plausibly an inline list, with at least three items, just abort
        if re.search(r'([;,]) \S+.*(\1|\1? and) \S', first_sentence) is None: return None 
        values = re.split(r'[;,] (?:and )?|[;,]? and ', first_sentence)

    return values

def excise_parentheticals(values):
    values = [re.sub(r'\s*[(][^)]+[)]', '', val) for val in values]
    return values

def resplit_list_values(values):
    values = filter(lambda val: re.match(r'No\s+', val, re.IGNORECASE) is None, values)
    values = [re.split(r'[;,] (?:and )?|[;,]? and | */ *(?=[a-zA-Z]{2})|\n', val) for val in values]
    values = [val.strip(' -.') for sublist in values for val in sublist]
    return values

def is_acceptable_value(value):
    if re.match(r'No\s+', value, re.IGNORECASE) is not None: return False
    if re.search(r'[a-z]', value, re.IGNORECASE) is None: return False
    return True

def cleanup_presenting_sx(llm_output_raw):
    values = split_listlike_text(llm_output_raw)
    if values is None: return None
    values = excise_parentheticals(values)
    values = resplit_list_values(values)
    # TODO: excise timing qualifiers (e.g., X days of, for X days, x N days)
    values = filter(is_acceptable_value, values)
    return "\n".join(values)


import numpy as np
import pandas as pd
import confidenceinterval as ci

from math import log10
from sklearn import metrics
from textwrap import dedent
from collections import OrderedDict

pd.set_option('future.no_silent_downcasting', True)

def filter_to_one_humans_rows_per_episode(df):
    """For each set of rows in a table, find the first human annotator per FK_episode_id,
    and only return their responses.
    
    Intended for use with the episode_labels table."""
    first_humans = df.groupby('FK_episode_id')['FK_human_annotator'].first().reset_index()

    return pd.merge(df, first_humans, on=['FK_episode_id', 'FK_human_annotator'], how='inner')


def create_vocab_matrix(df, vocab):
    """Creates a dataframe with a column for every term in `vocab`, and one row for each
    row in `df`. Any existing columns are preserved; any new columns are filled with False.
    The columns are sorted in alphabetical order."""
    cols_needed = [term for term in vocab.terms if term not in df]
    data = np.full((len(df), len(cols_needed)), False)
    df_empty = pd.DataFrame(data, columns=cols_needed, index=df.index)
    return pd.concat([df, df_empty], axis=1)[vocab.terms]


def f1_score_binary(y_true, y_pred):
    return ci.f1_score(y_true, y_pred, average='binary')


class ConfusionMatrix:
    # If these are a function, they should accept `y_true` and `y_pred` as arguments, similar to
    # the metrics in scikit-learn or the score functions in the `confidenceinterval` package
    # If these are a string, the corresponding method of the instance of this class is called
    #   (this allows instances to have attributes that affect the behavior of the metric)
    METRIC_METHODS = {
        "Accuracy": ci.accuracy_score,
        "Balanced accuracy": "_balanced_accuracy_score",
        "F1 score": f1_score_binary,
        "Precision aka PPV": ci.ppv_score,
        "Recall aka sensitivity": ci.tpr_score,
        "Specificity": ci.tnr_score,
        "NPV": ci.npv_score,
        "Cohen's kappa": "_cohen_kappa_score"
    }
    DEFAULT_N_RESAMPLES = 1000
    METRIC_PRECISION_SCALING = 2.4
    METRIC_MIN_DIGITS = 3

    def __init__(self, vec_truth, vec_pred, other_human=None, n_resamples=None):
        self._vec_truth = vec_truth
        self._vec_pred = vec_pred
        self._matrix = metrics.confusion_matrix(vec_truth, vec_pred)
        self._tn, self._fp, self._fn, self._tp = self._matrix.ravel()
        self._other_human = other_human

        # For metrics with bootstrapped confidence intervals
        self._n_resamples = self.DEFAULT_N_RESAMPLES if n_resamples is None else n_resamples
        self._random_generator = np.random.default_rng()
    
    @property
    def matrix(self): return self._matrix

    @property
    def tn(self): return self._tn

    @property
    def tp(self): return self._tp

    @property
    def fn(self): return self._fn

    @property
    def fp(self): return self._fp

    @classmethod
    def from_episode_labels(cls, df_truth, df_pred, vocab, max_line_num=9, **kwargs):
        df_truth = filter_to_one_humans_rows_per_episode(df_truth)

        df_truth = df_truth[["FK_episode_id", "label_name", "line_number"]].copy()
        df_truth["labeled"] = True
        df_truth_pivot = df_truth.pivot(index="FK_episode_id", columns="label_name", 
            values="labeled")
        df_truth_pivot.fillna(False, inplace=True)
        df_truth_mat = create_vocab_matrix(df_truth_pivot, vocab)

        df_pred = df_pred[["FK_episode_id", "label_name", "line_number"]].copy()
        df_pred["labeled"] = df_pred["line_number"] <= max_line_num
        df_pred_pivot = df_pred.pivot(index="FK_episode_id", columns="label_name", 
            values="labeled")
        df_truth_ep_ids = df_truth_pivot.reset_index()[["FK_episode_id"]]
        df_pred_pivot = pd.merge(df_truth_ep_ids, df_pred_pivot, 
            on="FK_episode_id", how="left", suffixes=('_X', ''))
        df_pred_pivot.fillna(False, inplace=True)
        df_pred_pivot.index = df_truth_ep_ids["FK_episode_id"]
        df_pred_mat = create_vocab_matrix(df_pred_pivot, vocab)

        vec_truth = df_truth_mat.values.flatten().astype(bool)
        vec_pred = df_pred_mat.values.flatten().astype(bool)
        return cls(vec_truth, vec_pred, **kwargs)


    def _bootstrapped_metric(self, y_true, y_pred, metric):
        return ci.bootstrap.bootstrap_ci(
            y_true=y_true,
            y_pred=y_pred,
            metric=metric,
            confidence_level=0.95,
            n_resamples=self._n_resamples,
            method='bootstrap_percentile',
            random_state=self._random_generator
        )


    def _balanced_accuracy_score(self, y_true, y_pred):
        return self._bootstrapped_metric(y_true, y_pred, metrics.balanced_accuracy_score)


    def _cohen_kappa_score(self, y_true, y_pred):
        return self._bootstrapped_metric(y_true, y_pred, metrics.cohen_kappa_score)


    def metrics(self):
        ret = OrderedDict()
        for metric, method in self.METRIC_METHODS.items():
            # Most metric methods are stateless functions, but some are dynamically obtained from
            #   this instance in order to capture external parameters (e.g., n_resamples)
            if not callable(method): method = getattr(self, method)
            # Metric methods expect two vectors, one for ground-truth and one for predictions
            # We now use methods from https://github.com/jacobgil/confidenceinterval
            # These are similar to sklearn.metrics but also provide a confidence interval 
            #   along with the point estimate
            ret[metric] = method(self._vec_truth, self._vec_pred)
        return ret
    

    def _format_metric(self, metric_name, metric_value):
        if isinstance(metric_value, tuple):
            prec = int(-log10(1 - metric_value[0]) + self.METRIC_PRECISION_SCALING)
            prec = max(prec, self.METRIC_MIN_DIGITS)
            # Metric includes a confidence interval
            metric_and_ci = (f"{metric_value[0]:10.{prec}f} ({metric_value[1][0]:9.{prec}f}, "
                f"{metric_value[1][1]:9.{prec}f})")
            return f"{metric_name:23.23} {metric_and_ci}"
        else:
            return f"{metric_name:23.23} {metric_value:10.2f}"


    def __str__(self):
        metrics = ("--------------------------- Metric ( 95% CI ) -------------\n" +
            "\n".join([self._format_metric(key, val) for key, val in self.metrics().items()]))

        header_1 = "                           Ground truth labels"
        header_2 = "                           Present     Absent"
        whose = "Predicted" if self._other_human is None else f"{self._other_human}'s"
        conf_mat = dedent(f"""\
            {header_1}
            {header_2}
            {whose+" label present":23} {str(self.tp):>10} {str(self.fp):>10}
            {whose+" label absent":23} {str(self.fn):>10} {str(self.tn):>10}""") 
        
        return conf_mat + "\n" + metrics


from os import path
from glob import glob
from importlib import import_module
from inspect import getmembers

from .abstract import TaskOutputType, AbstractTask

class TaskFinder:
    _index = None
    GLOB = '*.py'
    BASE_PATH = path.dirname(__file__)

    def _create_index(self):
        """Creates an index of the modules/classes within this package
        for use by the `find_tasks` method below."""
        self._index = []
        already_imported = set()

        for py_file in glob(path.join(self.BASE_PATH, self.GLOB)):
            if path.basename(py_file).startswith('__'): continue
            if path.basename(py_file).startswith('abstract'): continue

            py_file = py_file[len(self.BASE_PATH):]
            module_name = path.splitext(py_file)[0].replace(path.sep, '.')
            mod = import_module(module_name, package=__name__)

            for name, obj in getmembers(mod):
                if name.startswith('_') or name.startswith('Abstract'): continue
                if obj in already_imported: continue
                if not isinstance(obj, type) or not issubclass(obj, AbstractTask):
                    continue

                already_imported.add(obj)
                self._index.append({
                    "output_type": obj.output_type,
                    "task": obj, 
                    "task_name": obj.name,
                    "class_name": name,
                })

    # Lazily create the indexes, once, upon first access
    @property
    def index(self):
        if self._index is not None: return self._index
        self._create_index()
        return self._index
    
    def find_tasks(self, name=None, output_type=None):
        tasks = []

        for row in self.index:
            if name is not None and row["task_name"] != name: continue

            if output_type is not None:
                if isinstance(output_type, str):
                    if row["output_type"].value != output_type: continue
                else:
                    if not isinstance(output_type, TaskOutputType):
                        raise RuntimeError("output_type must be `str` or `TaskOutputType`")
                    if output_type is not row["output_type"]: continue

            tasks.append((row["task_name"], row["class_name"], row["task"]))

        tasks = [row[2] for row in sorted(tasks, key = lambda row: row[0:1])]
        return tasks


# We create one global instance of the above class
_finder = TaskFinder()

# Export this prebound method
find_tasks = _finder.find_tasks


"""Pytest configuration and shared fixtures."""
import os
import shutil
import subprocess
import uuid
import warnings
from contextlib import contextmanager
from pathlib import Path
import pytest
from llacie.config import Config


@contextmanager
def _create_temp_database():
    """Context manager to create and cleanup a temporary PostgreSQL database.

    Yields the llacie Config object for the test database.
    """
    db_name = f"llacie_test_{uuid.uuid4().hex[:8]}"
    old_config = Config()

    # Store original environment variables
    original_db_name = os.environ.get("PG_DBNAME")
    original_pgpassword = os.environ.get("PGPASSWORD")
    os.environ["PGPASSWORD"] = old_config["PG_PASS"]
    cmd_args = ["-h", old_config["PG_HOST"], "-U", old_config["PG_USER"]]

    # Create the database
    try:
        subprocess.run(
            ["createdb", *cmd_args, db_name],
            check=True,
            capture_output=True,
            text=True
        )
    except subprocess.CalledProcessError as e:
        raise Exception(f"Could not create test database: {e.stderr}")

    # Set PG_DBNAME to the test database
    os.environ["PG_DBNAME"] = db_name
    new_config = Config()

    try:
        yield new_config
    finally:
        # Cleanup: restore original PG_DBNAME
        if original_db_name:
            os.environ["PG_DBNAME"] = original_db_name
        else:
            os.environ.pop("PG_DBNAME", None)

        # Drop the test database
        try:
            subprocess.run(
                ["dropdb", *cmd_args, "--if-exists", db_name],
                check=True,
                capture_output=True,
                text=True
            )
        except subprocess.CalledProcessError:
            warnings.warn(f"Failed to cleanup test Postgres database '{db_name}'", UserWarning)
        finally:
            # Restore original PGPASSWORD
            if original_pgpassword:
                os.environ["PGPASSWORD"] = original_pgpassword
            else:
                os.environ.pop("PGPASSWORD", None)


def _setup_isolated_env(tmp_path):
    """Set up an isolated test environment with examples and config files.

    Args:
        tmp_path: Temporary directory path

    Returns:
        Path to the temporary directory
    """
    # Copy test data
    test_data = Path(__file__).parent.parent / "examples"
    if test_data.exists():
        shutil.copytree(test_data, tmp_path / "examples")

    # Copy dotenv files
    for config_file in [".env.example", ".env"]:
        config_file = Path(__file__).parent.parent / config_file
        if config_file.exists():
            shutil.copy(config_file, tmp_path)

    return tmp_path


@pytest.fixture
def temp_db():
    """Create a temporary PostgreSQL test database (function-scoped).

    Uses existing PG_HOST, PG_USER, and PG_PASS environment variables
    and creates a uniquely named test database.
    """
    with _create_temp_database() as config:
        yield config


@pytest.fixture
def isolated_env(tmp_path, monkeypatch):
    """Create an isolated environment for testing (function-scoped)."""
    monkeypatch.chdir(tmp_path)
    return _setup_isolated_env(tmp_path)


@pytest.fixture(scope="class")
def class_temp_db():
    """Class-scoped temporary PostgreSQL database.

    Shared across all tests in a test class to avoid recomputing expensive operations.
    """
    with _create_temp_database() as config:
        yield config


@pytest.fixture(scope="class")
def class_isolated_env(tmp_path_factory):
    """Class-scoped isolated environment.

    Shared across all tests in a test class to avoid recomputing expensive operations.
    Changes to the temporary directory for the duration of the test class.
    """
    tmp_path = tmp_path_factory.mktemp("llacie_class")
    original_cwd = os.getcwd()

    # Setup environment and change directory
    os.chdir(tmp_path)
    result = _setup_isolated_env(tmp_path)

    yield result

    # Restore original directory
    os.chdir(original_cwd)



import sys
import json

from os import path
from textwrap import dedent
from socket import gethostname
from glob import glob

from .abstract import AbstractStrategy
from ..batch.apptainer_slurm import ApptainerSlurmJobManager
from ..cache.sqlite import SqliteFeatureWorkerCache
from ..inference.vllm import Vllm
from ..inference.llama_cpp_python import LlamaCppPython
from ..utils import echo_warn, echo_info, strtobool

class AbstractVllmOrLcpStrategy(AbstractStrategy):
    """
    Abstract class for running vllm or llama-cpp-python with constant parameters and a prespecified 
    prompt template, across all of the specified notes, substituting one section for each note into
    the prompt template to extract a feature. The backend (vllm vs llama-cpp-python) is configurable
    and all options are translated to match either backend wherever possible.

    The output is currently constrained to be a JSON array of strings. This could be overridden
    by inheriting classes, by redefining `self.JSON_SCHEMA`.
    
    Slurm may be used to schedule the jobs. On ERISXdl, we are required to use both slurm and 
    Apptainer containers to access the GPUs on the GPU nodes.

    This class is intended to be subclassed and configured *primarily* by overriding the class 
    constants, and beyond that, its methods might be overridden as well.
    """
    # can be, e.g., ShortHPISectionRegexStrategy from .section.hpi_short.regex
    SECTION_STRATEGY_CLASS = None             

    # These are example default options based on the Llama 3 8B strategy for presenting symptoms
    # This is a HuggingFace respository ID for the full PyTorch/Transformers model, which vllm uses
    MODEL_ID = 'meta-llama/Meta-Llama-3-8B-Instruct'
    # As a fallback, you can also provide a GGUF version, which is what llama-cpp-python uses
    GGUF_MODEL_ID = 'QuantFactory/Meta-Llama-3-8B-Instruct-GGUF'
    GGUF_MODEL_FILENAME = '*.Q5_K_S.gguf'

    SLURM_ENABLED = True
    SLURM_THREADS_PER_JOB = 10
    ARRAY_OF_STRINGS_JSON_SCHEMA = """
    {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "array",
        "items": {
            "type": "string"
        }
    }
    """
    JSON_SCHEMA = ARRAY_OF_STRINGS_JSON_SCHEMA

    # This flag and config var allows this class to toggle between vllm and llama-cpp-python backends
    VLLM_ENABLED = False
    # Need to set vllm's LLM(..., dtype='half') because the Nvidia V100s in the ERISXdl cluster do 
    # not natively support bfloat16. 
    # This is expected to trigger a warning: "Casting torch.bfloat16 to torch.float16"
    VLLM_DTYPE = 'half'

    # The maximum number of output tokens you'd ever expect to generate per prompt
    LLM_TOKENS_TO_GENERATE = 256
    # Arguments to SamplingParams in vllm
    # See: https://docs.vllm.ai/en/latest/api/vllm/#vllm.SamplingParams
    # These are translated into equivalent arguments for llama-cpp-python's create_chat_completion
    # https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.Llama.create_chat_completion
    SAMPLING_PARAMS = {
        "temperature": 0.1,
        "top_p": 0.8,
        "repetition_penalty": 1.05
    }

    NICE_MODEL_NAME = "Llama3 8B Instruct"

    # Model context length. Note vllm cannot generate past the end of the context window.
    LLM_MAX_MODEL_LENGTH = 2048  

    # Prompts
    LLM_SYSTEM_PROMPT = dedent("""\
        You are a clinical researcher that reads medical charts and answers questions about them. 
        Use only the information in the text provided to answer the question.
        If a patient denies something, do not include it in your answer.
        After you provide an answer, you immediately stop talking.""")
    LLM_USER_PROMPT = dedent("""\
        Read the following patient history and list the patient's presenting symptoms.
        Include only symptoms present now or reported for the days to weeks leading up to admission.
        Ignore any symptoms from past medical history or prior hospital admissions.
        Give your answer as a JSON array containing up to ten strings.
        Each string contains between one and three words.
        
        {input}""")


    def __init__(self, db, config, **options):
        super().__init__(db, config, **options)
        # Ensure this config option exists because ApptainerSlurmJobManager's parent class needs it
        self.config['SLURM_THREADS_PER_JOB'] = config.get("SLURM_THREADS_PER_JOB",
            self.SLURM_THREADS_PER_JOB)
        vllm_enabled = strtobool(config.get("VLLM_ENABLED", self.VLLM_ENABLED))
        self._dtype = config.get("VLLM_DTYPE", self.VLLM_DTYPE)
        self.backend = "vllm" if vllm_enabled else "llama-cpp-python"
        
        if not isinstance(self.SECTION_STRATEGY_CLASS, type):
            raise NotImplementedError("You must specify a SECTION_STRATEGY_CLASS")

        self.sec_strat = self.SECTION_STRATEGY_CLASS(db, config, **options)
        self.sec_task_name = self.sec_strat.task.name

        self._set_model_paths(config)

        self._sampling_params = self._translate_sampling_params({
            **self.SAMPLING_PARAMS, 
            "max_tokens": self.LLM_TOKENS_TO_GENERATE
        })
        self._engine_args = self._create_engine_args()

        self.slurm_enabled = strtobool(config.get("SLURM_ENABLED", self.SLURM_ENABLED))
        self.workers_have_networking = config.get("APPTAINER_WORKERS_HAVE_NETWORKING", 
            ApptainerSlurmJobManager.APPTAINER_WORKERS_HAVE_NETWORKING)

        if self.slurm_enabled and not self.workers_have_networking:
            self.worker_cache = SqliteFeatureWorkerCache(db, config, self)


    def _set_model_paths(self, config):
        self.local_models_path = config.get("LOCAL_MODELS_PATH", None)
        self.model_path = self.MODEL_ID if self.backend == "vllm" else self.GGUF_MODEL_ID
        self.tokenizer_model_path = self.MODEL_ID
        if self.local_models_path: 
            self.model_path = f"{self.local_models_path}/{self.model_path}"
            self.tokenizer_model_path = f"{self.local_models_path}/{self.tokenizer_model_path}"
            if self.backend == "llama-cpp-python":
                self.model_path = f"{self.model_path}/{self.GGUF_MODEL_FILENAME}"
                try: self.model_path = glob(self.model_path)[0]
                except IndexError: raise RuntimeError(f"Could not find GGUF: {self.model_path}")


    def _translate_sampling_params(self, params):
        """Converts sampling params from vllm's SamplingParams keys to the arguments used
        by create_chat_completion in llama-cpp-python. Most are exactly the same, but this fixes
        the few key exceptions. See:
        https://docs.vllm.ai/en/latest/api/vllm/#vllm.SamplingParams
        https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.Llama.create_chat_completion"""
        if self.backend == "vllm": return params
        if "repetition_penalty" in params: 
            params["repeat_penalty"] = params.pop("repetition_penalty")
        return params


    def _create_engine_args(self):
        args = {}
        if self.backend == "vllm":
            args["dtype"] = self._dtype
            args["max_model_len"] = self.LLM_MAX_MODEL_LENGTH
        else: # self.backend == "llama-cpp-python"
            args["tokenizer_model_path"] = self.tokenizer_model_path
            args["verbose"] = False
            args["n_ctx"] = self.LLM_MAX_MODEL_LENGTH
            if not self.local_models_path: 
                args["filename"] = self.GGUF_MODEL_FILENAME
        return args


    def create_engine(self):
        engine = Vllm if self.backend == "vllm" else LlamaCppPython
        return engine(self.config, self.model_path, self.LLM_SYSTEM_PROMPT, self.LLM_USER_PROMPT,
            self.JSON_SCHEMA, **self._engine_args)


    def _run_llm(self, note_ids):
        if self.within_worker and not self.workers_have_networking: db = self.worker_cache
        else: db = self.db

        llm = self.create_engine()
        max_prompt_length = self.LLM_MAX_MODEL_LENGTH - self.LLM_TOKENS_TO_GENERATE - 4

        self.start_timer()

        prompts = []
        section_ids = []
        
        for note_id in note_ids:
            # Get the note section to work on
            row = db.get_note_section(note_id, self.sec_strat)
            if row is None: 
                raise RuntimeError(f"No {self.sec_task_name} section found for note id {note_id}")
            section_value = row.section_value

            # Ensure the prompt will fit in the context window, with 4 tokens of wiggle room
            full_prompt = llm.autotrim_prompt(max_prompt_length, input=section_value)
            if full_prompt is None:
                echo_warn(f"Autotrimming failed for note id {note_id}")
                # Couldn't autotrim; likely a malformed section, but try using the whole thing
                full_prompt = llm.create_prompt(input=section_value)
            
            section_ids.append(row.id)
            prompts.append(full_prompt)
            
        # Run all the prompts through the LLM backend
        # Outputs are already deserialized from JSON into Python lists or dicts
        echo_info(f"Running {len(note_ids)} notes thru the {self.NICE_MODEL_NAME} model "
            f"using the {self.backend} backend")
        outputs = llm(prompts, **self._sampling_params)
        runtime = self.stop_timer() / len(outputs)

        for i, output in enumerate(outputs):
            note_id = note_ids[i]
            section_id = section_ids[i]
            if output is not None and isinstance(output, list): 
                feature = "\n".join(output)
                output_raw = json.dumps(output)
                db.upsert_note_feature(note_id, self, output_raw, feature, section_id, runtime)
            else:
                echo_warn(f"No LLM output generated for note id {note_id}")


    def run(self, all_note_ids, max_note_ids=None, batch_size=None, dry_run=False):
        if self.within_worker:
            # We are within a worker process; run the LLM on the given chunk of note IDs
            if dry_run: return echo_info("Info: Dry run, exiting without running any LLMs")
            note_ids = all_note_ids
            echo_info(f"Worker started on {gethostname()}") 

        else:
            # There is no point running this on notes without the required section
            note_ids = self.db.filter_to_notes_with_section(all_note_ids, self.sec_strat)
            if (num_filtered := len(all_note_ids) - len(note_ids)) > 0:
                echo_warn(f"Skipping {num_filtered} notes without an {self.sec_task_name} section")
            if max_note_ids is not None and (len(note_ids) > max_note_ids):
                echo_warn(f"Limiting to the first {max_note_ids} notes")
                note_ids = note_ids[:max_note_ids]
            if dry_run: return echo_info("Info: Dry run, exiting without running any LLMs")
            
            if self.slurm_enabled:
                # Create a SlurmJobManager and submit jobs that run worker processes
                worker_cmd = sys.argv.copy()
                # We want a relative path to the `llacie` executable, because the worker processes
                #   may be running in a different virtualenv than this one
                worker_cmd[0] = path.basename(worker_cmd[0])
                # To start a worker, you invoke the same command with `--worker` appended
                worker_cmd.append("--worker")
                job_manager = ApptainerSlurmJobManager(worker_cmd, self.config, self.worker_cache)
                return job_manager.run(note_ids)
        
        # Either slurm is disabled or we are inside one the worker processes
        # In this case, we should start running vLLM on the note IDs
        self._run_llm(note_ids)

        # For worker processes without networking, we need to mark the results on disk as complete
        if self.within_worker and not self.workers_have_networking: self.worker_cache.finalize()


from . import AbstractPresentingSymptomsEpisodeLabelV2Strategy
from ...feature.presenting_sx.llama3_8b import PresentingSxFeatureLlama3Instruct8BStrategy
from ....tasks.episode_label import PresentingSymptomsEpisodeLabelV2Task

class PresentingSymptomsEpisodeLabelV2Llama3Instruct8BVllmStrategy(
        AbstractPresentingSymptomsEpisodeLabelV2Strategy):
    """\
    Converts the ..presenting_sx.llama3_8b_vllm note features into episode labels using vocab v2.
    This vocab is UPDATED since the K08 submission after manual examination of the
    misclassifications.
    """
    task = PresentingSymptomsEpisodeLabelV2Task
    name = "episode_label.pres_sx_eplab2.llama3_8b"
    version = "0.0.1"
    prereq_tasks = []

    BATCH_SIZE = 1000


    def __init__(self, db, config, **options):
        super().__init__(db, config, **options)
        self.feat_strat = PresentingSxFeatureLlama3Instruct8BStrategy(db, config, **options)


from .abstract import AbstractFeatureTask

class PresentingSymptomsFeatureTask(AbstractFeatureTask):
    """\
    Presenting symptoms found in the `hpi_short` section of a note.
    
    Limited to H&P notes only, for patients with suspected infection,
    and applying the exclusions listed in Pak et al. CID 2023:
    https://pubmed.ncbi.nlm.nih.gov/37531612/
    """
    name = "presenting_sx"
    note_join_sql = """
        LEFT JOIN "{{cohort_table}}" AS esc ON
            n."FK_episode_id" = esc."FK_episode_id"
    """
    note_where_sql = """
        "InpatientNoteTypeDSC" IN ('H&P')
        AND "infectionCriteria" IS TRUE
        AND "excl_ST0_combined" IS FALSE
    """


import json
from textwrap import dedent
from collections import defaultdict

from ...abstract_vllm_or_lcp import AbstractVllmOrLcpStrategy
from ...section.hpi_short.regex import ShortHPISectionRegexStrategy
from ....tasks.feature import PresentingSymptomsFeatureTask
from ....vocab import Vocab
from ....utils import chunker, echo_info, echo_warn


class PresentingSxFeatureLlama3Instruct8BByTermStrategy(AbstractVllmOrLcpStrategy):
    """\
    Attempts to extract the `presenting_sx` feature using Llama 3 Instruct 8B model,
    an 8B parameter model in the Llama family, and with the prompt including terms in the vocab.

    This is modified from the `feature.presenting_sx.llama3_8b` strategy by using multiple 
    prompts to ask about every symptom in a vocab, ~5 terms at a time. The downside is that many 
    more prompts (and tokens) are required per note, but there is a potential upside in accuracy
    since the task given to the LLM is more constrained and .
    """
    task = PresentingSymptomsFeatureTask
    name = "feature.presenting_sx.llama3_8b_byterm"
    version = "0.0.1"
    prereq_tasks = []

    SECTION_STRATEGY_CLASS = ShortHPISectionRegexStrategy

    # HuggingFace respository ID for the full PyTorch/Transformers model, which vllm uses
    MODEL_ID = 'meta-llama/Meta-Llama-3-8B-Instruct'
    # As a fallback, you can also provide a GGUF version, which is what llama-cpp-python uses
    GGUF_MODEL_ID = 'QuantFactory/Meta-Llama-3-8B-Instruct-GGUF'
    GGUF_MODEL_FILENAME = '*.Q5_K_S.gguf'
    
    SLURM_IDS_PER_JOB = 50
    SLURM_THREADS_PER_JOB = 16
    OBJECT_WITH_FIVE_PROPERTIES_JSON_SCHEMA = """
    {
        "$schema": "http://json-schema.org/draft-07/schema#",
        "type": "object",
        "additionalProperties": {
            "type": "boolean"
        },
        "minProperties": 1,
        "maxProperties": 5
    }
    """
    JSON_SCHEMA = OBJECT_WITH_FIVE_PROPERTIES_JSON_SCHEMA

    VLLM_ENABLED = False
    VLLM_DTYPE = 'half'
    LLM_TOKENS_TO_GENERATE = 1024
    # Arguments to SamplingParams in vllm
    # See: https://docs.vllm.ai/en/latest/api/vllm/#vllm.SamplingParams
    # These are translated into equivalent arguments for llama-cpp-python's create_chat_completion
    SAMPLING_PARAMS = {
        "temperature": 0.1,
        "top_p": 0.8,
        "repetition_penalty": 1.05
    }

    NICE_MODEL_NAME = "Llama3 8B Instruct **ByTerm**"

    # Model context length. Note that Vllm will not generate past the end of the context window.
    LLM_MAX_MODEL_LENGTH = 3044  
    LLM_SYSTEM_PROMPT = dedent("""\
        You are a clinical researcher that reads medical charts and answers questions about them. 
        Use only the information in the text provided to answer the question.
        After you provide an answer, you immediately stop talking.""")
    LLM_USER_PROMPT = dedent("""\
        Read the following patient history and figure out which of the following presenting symptoms
        are reported for this patient: {symptoms}.
        Include only symptoms present now or reported for the days to weeks leading up to admission.
        Ignore any symptoms from past medical history or prior hospital admissions.
        Give your answer as a JSON object with the symptoms as the keys and true/false as the 
        values. For example, if all symptoms are absent, the answer would be: {symptoms_absent_json}
        
        Now, here is the patient history:
        
        {input}""")
    
    # Specify a vocab to draw terms from, and how many terms to use at a time in the prompt
    VOCAB_XLSX = "dt.pres_sx.ngrams.v2.xlsx"
    VOCAB_SHEET_NAME = "symptom_ngrams_top75ile.edited"
    TERMS_PER_PROMPT = 5


    def __init__(self, db, config, **options):
        super().__init__(db, config, **options)
        # We decrease the default number of note IDs per Slurm job, since each note now generates
        # hundreds of prompts to the LLM (rather than one each)
        self.config['SLURM_IDS_PER_JOB'] = config.get("SLURM_IDS_PER_JOB",
            self.SLURM_IDS_PER_JOB)


    def _run_llm(self, note_ids):
        if self.within_worker and not self.workers_have_networking: db = self.worker_cache
        else: db = self.db

        llm = self.create_engine()
        max_prompt_length = self.LLM_MAX_MODEL_LENGTH - self.LLM_TOKENS_TO_GENERATE - 4
        vocab = Vocab(self.VOCAB_XLSX, sheet_name=self.VOCAB_SHEET_NAME)

        self.start_timer()

        prompts = []
        section_ids = {}
        out_note_ids = []

        for note_id in note_ids:
            # Get the note section to work on
            row = db.get_note_section(note_id, self.sec_strat)
            if row is None: 
                raise RuntimeError(f"No {self.sec_task_name} section found for note id {note_id}")
            section_value = row.section_value
            section_ids[note_id] = row.id

            for terms in chunker(vocab.terms, self.TERMS_PER_PROMPT):
                prompt_args = {
                    "input": section_value,
                    "symptoms": ", ".join(terms),
                    "symptoms_absent_json": json.dumps({term: False for term in terms})
                }

                # Ensure the prompt will fit in the context window, with 4 tokens of wiggle room
                full_prompt = llm.autotrim_prompt(max_prompt_length, **prompt_args)
                if full_prompt is None:
                    echo_warn(f"Autotrimming failed for note id {note_id}")
                    # Couldn't autotrim; likely a malformed section, but try using the whole thing
                    full_prompt = llm.create_prompt(**prompt_args)
                
                out_note_ids.append(note_id)
                prompts.append(full_prompt)
            
        # Run all the prompts through the LLM backend
        # Outputs are already deserialized from JSON into Python lists or dicts
        echo_info(f"Running {len(note_ids)} notes thru the {self.NICE_MODEL_NAME} model "
            f"using the {self.backend} backend")
        outputs = llm(prompts, **self._sampling_params)
        runtime = self.stop_timer() / len(note_ids)

        outputs_by_note_id = defaultdict(dict)
        for i, output in enumerate(outputs):
            note_id = out_note_ids[i]
            if output is not None and isinstance(output, dict): 
                outputs_by_note_id[note_id].update(output)
            else:
                echo_warn(f"Invalid LLM output generated for note id {note_id}")
        
        for note_id, output in outputs_by_note_id.items():
            present_terms = [term for term, presence in output.items() if presence is True]
            feature = "\n".join(present_terms)
            output_raw = json.dumps(output)
            section_id = section_ids[note_id]
            db.upsert_note_feature(note_id, self, output_raw, feature, section_id, runtime)




import os
import click
import collections
from ..app import App
from ..utils import echo_err

class OrderedGroup(click.Group):
    def __init__(self, name=None, commands=None, **attrs):
        super(OrderedGroup, self).__init__(name, commands, **attrs)
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx):
        return self.commands

@click.group(cls=OrderedGroup)
@click.pass_context
def main(ctx):
    """LLaCIE: Large Language model Clinical Information Extraction
    
    This is the main menu. Please use one of the subcommands listed below."""
    ctx.obj = ctx.with_resource(App())

@main.command()
@click.option('--overwrite/--no-overwrite', default=False,
    help="Forces the dropping and re-creation of tables if they already exist.")
@click.option('--missing-tables-only', default=False, is_flag=True,
    help="Create only missing tables, if possible; do NOT drop any existing tables.")
@click.pass_obj
def init_db(app, overwrite, missing_tables_only):
    """(0) Creates the database structure used by LLaCIE."""
    app.db.create_tables(overwrite, missing_tables_only)

@main.group(cls=OrderedGroup)
@click.pass_obj
def import_notes(app):
    """(1) Import notes from various sources."""
    pass

@import_notes.command(name="text")
@click.argument('input_path', type=click.Path(exists=True))
@click.option('--overwrite/--no-overwrite', default=False,
    help="Permits the `notes` table to be truncated if it already contains rows.")
@click.pass_obj
def import_notes_text(app, input_path, overwrite):
    """Import notes from plain text files."""
    app.import_notes_text(input_path, overwrite)

@import_notes.command(name="edw")
@click.option('--overwrite/--no-overwrite', default=False,
    help="Permits the `notes` table to be truncated if it already contains rows.")
@click.option('--resume-note-text', default=False, is_flag=True,
    help="Resumes downloading of note text only, without rebuilding the table.")
@click.pass_obj
def import_notes_edw(app, overwrite, resume_note_text):
    """Import notes from the MGB EDW (Enterprise Data Warehouse)."""
    app.import_notes_edw(overwrite, resume_note_text)

@main.group(cls=OrderedGroup)
@click.pass_obj
def sections(app):
    """(2) Extract sections from each note."""
    pass

@sections.command(name="list")
@click.option('-n', '--section-name', default=None, type=str, 
    help="Filter to only strategies for one section name.")
@click.option('-s', '--strategy-name', default=None, type=str, 
    help="Filter to strategies matching this name. Partial matches using shell-style globs are allowed.")
@click.pass_obj
def list_sections(app, section_name=None, strategy_name=None):
    """List available sections, along with strategies for each."""
    app.list_sections_with_strategies(section_name, strategy_name)

@sections.command(name="extract")
@click.option('-n', '--section-name', default=None, type=str, 
    help="Specifies which section to extract.")
@click.option('-s', '--strategy-name', default=None, type=str, 
    help="Specifies which strategy to run. Partial matches using shell-style globs are allowed.")
@click.option('-d', '--dry-run', default=False, is_flag=True, 
    help="Prints how many notes would be processed and exits; the strategy is not run.")
@click.pass_obj
def extract_sections(app, section_name, strategy_name, dry_run):
    """Extract a section using a specified strategy."""
    app.run_section_strategy(section_name, strategy_name, dry_run)


@main.group(cls=OrderedGroup)
@click.pass_obj
def features(app):
    """(3) Extract features from each note."""
    pass

@features.command(name="list")
@click.option('-n', '--feature-name', default=None, type=str, 
    help="Filter to only strategies for one feature name.")
@click.option('-s', '--strategy-name', default=None, type=str, 
    help="Filter to strategies matching this name. Partial matches using shell-style globs are allowed.")
@click.pass_obj
def list_features(app, feature_name=None, strategy_name=None):
    """List available features, along with strategies for each."""
    app.list_features_with_strategies(feature_name, strategy_name)

@features.command(name="extract")
@click.option('-n', '--feature-name', default=None, type=str, 
    help="Filter to only strategies for one feature name.")
@click.option('-s', '--strategy-name', default=None, type=str, 
    help="Filter to strategies matching this name. Partial matches using shell-style globs are allowed.")
@click.option('-d', '--dry-run', default=False, is_flag=True, 
    help="Prints how many notes would be processed and exits; the strategy is not run.")
@click.option('-A', '--annotated-for-task', default=None, type=str, 
    help="Run only for notes that have human annotations for the given task name.")
@click.option('-M', '--max-note-ids', default=None, type=int, 
    help="Maximum number of notes to process in this invocation.")
@click.option('-R', '--redo-older-than', default=None, type=click.DateTime(), 
    help=("Any features extracted before this time will be considered stale "
        "and re-extracted."))
@click.option('--worker', is_flag=True, hidden=True, 
    help="For internal use only, used to run this subcommand within a bsub job.")
@click.argument('note_ids', type=int, nargs=-1)
@click.pass_obj
def extract_features(app, feature_name=None, strategy_name=None, dry_run=False, 
        annotated_for_task=None, max_note_ids=None, worker=False, redo_older_than=None, 
        note_ids=tuple()):
    """Extract features from notes using a specified strategy.
    
    The NOTE_IDS argument is for internal use only, when this is called as a worker 
    process, to run the strategy on specific rows in the database."""
    if worker:
        app.is_worker = True
        annotated_for_task = None
    elif len(note_ids) > 0:
        return echo_err("Error: Not allowed to specify NOTE_IDS when invoking this manually.")
    app.run_feature_strategy(
        feature_name, 
        strategy_name, 
        dry_run,
        annotated_for_task, 
        max_note_ids,
        redo_older_than,
        note_ids
    )


@main.group(cls=OrderedGroup)
@click.pass_obj
def episode_labels(app):
    """(4) Apply labels to episodes of care using the features."""
    pass

@episode_labels.command(name="list")
@click.option('-n', '--episode-label-task', default=None, type=str, 
    help="Filter to only strategies for one episode labelling task.")
@click.option('-s', '--strategy-name', default=None, type=str, 
    help="Filter to strategies matching this name. Partial matches using shell-style globs are allowed.")
@click.pass_obj
def list_episode_labels(app, episode_label_task=None, strategy_name=None):
    """List available episode labels, along with strategies for each."""
    app.list_episode_labels_with_strategies(episode_label_task, strategy_name)

@episode_labels.command(name="extract")
@click.option('-n', '--episode-label-task', default=None, type=str, 
    help="Filter to only strategies for episode labelling task name.")
@click.option('-s', '--strategy-name', default=None, type=str, 
    help="Filter to strategies matching this name. Partial matches using shell-style globs are allowed.")
@click.option('-d', '--dry-run', default=False, is_flag=True, 
    help="Prints how many episodes would be processed and exits; the strategy is not run.")
@click.option('-A', '--annotated-for-task', default=None, type=str, 
    help="Run only for episodes that have human annotations for the given task name.")
@click.pass_obj
def extract_episode_labels(app, episode_label_task=None, strategy_name=None, dry_run=False,
        annotated_for_task=None):
    """Create episode labels using a specified strategy."""
    app.run_episode_label_strategy(episode_label_task, strategy_name, dry_run, annotated_for_task)

@episode_labels.command(name="import")
@click.argument('episode_label_task', type=str)
@click.argument('input_xlsx', type=click.File('rb'))
@click.option('-w', '--sheet-name', default=None, type=str, 
    help="Specify the name of the worksheet in the XLSX that will be imported. "
    "Defaults to the first sheet.")
@click.option('-u', '--human-username', default=os.getenv('USER'), type=str, 
    help="Specify the username to whom imported annotations will be attributed. "
    "Defaults to the current user.")
@click.pass_obj
def import_episode_labels(app, episode_label_task, input_xlsx, sheet_name=None, 
        human_username=None):
    """Import human-annotated episode labels from an XLSX file.
    
    Expects an XLSX worksheet with a header row and two specific columns:

    - `FK_episode_id` - must correspond to a valid `id` in the episodes table

    - `human_labels` - the human-created label names, delimited by | (pipes)"""
    app.db.import_episode_labels(episode_label_task, input_xlsx, sheet_name, human_username)

@episode_labels.command(name="evaluate")
@click.option('-n', '--episode-label-task', default=None, type=str, 
    help="Filter to only strategies for this episode labelling task name.")
@click.option('-s', '--strategy-name', default=None, type=str, 
    help="Filter to strategies matching this name. Partial matches using shell-style globs are allowed.")
@click.option('-u', '--human-username', default=None, type=str, 
    help="Specify the username for whom all strategies' annotations will be compared. "
    "Defaults to the first available human annotation.")
@click.option('-o', '--other-human', default=None, type=str, 
    help="Specify another human's username, whose labels will be treated an alternate strategy, "
    "in effect evaluating inter-labeler reliability.")
@click.option('-l', '--also-labeled-by', default=None, type=str, 
    help="Specify another human's username, who must have labeled each of the episodes included "
    "for the evaluation. Can be different from -u.")
@click.option('-B', '--bootstrap-samples', default=1000, type=int, 
    help="For bootstrapped confidence intervals, specify the # of resamples. Default is 1000.")
@click.pass_obj
def evaluate_episode_labels(app, episode_label_task=None, strategy_name=None,
        human_username=None, other_human=None, also_labeled_by=None, bootstrap_samples=None):
    """Evaluate episode labelling strategies against human annotations."""
    app.evaluate_episode_label_strategies(episode_label_task, strategy_name, human_username,
        other_human, also_labeled_by, bootstrap_samples)


import os
import re
import getpass
import pandas as pd
import numpy as np

from collections import OrderedDict
from click import UsageError
from datetime import datetime
from sqlalchemy import create_engine, text, MetaData, Table
from sqlalchemy.engine.reflection import Inspector
from contextlib import contextmanager

from .config import ConfigError
from .tasks import TaskOutputType as out_t
from .tasks.abstract import AbstractTask
from .strategies import find_strategies
from .strategies.abstract import AbstractStrategy
from .utils import echo_info, echo_warn, generate_unique_id, create_row_like


HUNDRED_YEARS_IN_SECONDS = 60 * 60 * 24 * 365 * 100
LAST_UPDATED_COLUMNS = {
    out_t.SECTION: "last_updated",
    out_t.FEATURE: "feature_updated"
}

class LlacieDatabase(object):
    def __init__(self, config):
        if config['PG_URI'] is None:
            raise ConfigError("Could not build postgres connection string")

        self._conn = None
        self.config = config
        self.prefix = config.get('PG_TABLE_PREFIX', '')
        self.ep_table = config.get('PG_EPISODE_TABLE', 'llacie_episodes')
        self.cohort_table = config.get('PG_COHORT_TABLE', 'llacie_cohorts')
        self._metadata = None

    @property
    def conn(self):
        # Lazily initialize the connection to the database only when needed
        if self._conn is None:
            self._conn = create_engine(self.config['PG_URI']).connect()
        return self._conn

    @property
    def metadata(self):
        """A property that accesses reflected table MetaData created by SQLAlchemy."""
        if self._metadata is not None: return self._metadata
        self._metadata = MetaData()
        self._metadata.reflect(bind=self.conn)
        return self._metadata

    @property
    def tables(self):
        """A property that reflects all of the SQLAlchemy table objects.
        See: https://docs.sqlalchemy.org/en/20/core/reflection.html#reflecting-all-tables-at-once"""
        return self.metadata.tables
    
    def table(self, table_name):
        return Table(f"{self.prefix}{table_name}", self.metadata, autoload_with=self.conn)

    def interpolate_config_vars(self, sql):
        sql = re.sub('{{prefix}}', self.prefix, sql)
        sql = re.sub('{{episode_table}}', self.ep_table, sql)
        return re.sub('{{cohort_table}}', self.cohort_table, sql)

    def close(self):
        if self._conn is not None:
            self._conn.close()

    #############################
    # INITIALIZING THE DATABASE
    #############################

    def create_tables(self, overwrite = False, missing_tables_only = False):
        """Creates all of the database tables needed to run LLaCIE."""
        inspect = Inspector.from_engine(self.conn)

        sql_file = os.path.abspath(os.path.join(__file__, "../sql/schema.sql"))
        with open(sql_file) as f:
            sql = self.interpolate_config_vars(f.read())
            sql_chunks = re.split(r'^\s*CREATE TABLE\s+', sql, flags = re.MULTILINE)
            
            tbl_chunks = OrderedDict()
            llacie_tables = []
            for sql_chunk in sql_chunks:
                if len(sql_chunk.strip()) == 0: continue
                match = re.match(r'(IF +NOT +EXISTS +)?([a-zA-Z0-9_"]+) +[(]', sql_chunk)
                if match is None: 
                    raise ConfigError("Invalid schema in sql/schema.sql")

                table_name = match.group(2).strip('"')
                tbl_chunks[table_name] = f"CREATE TABLE {sql_chunk}"
                if match.group(1) is None: llacie_tables.append(table_name)

            existing_tables = [tbl for tbl in llacie_tables if inspect.has_table(tbl)]
            if not missing_tables_only and len(existing_tables) > 0:
                if not overwrite:
                    raise UsageError(f"Tables {', '.join(existing_tables)} already exist. If "
                        "you want to drop and create new blank tables, use --overwrite. "
                        "If you only want to add missing tables, use --missing-tables-only.")
                else:
                    # Existing tables are deleted in reverse order of creation
                    for existing_table in reversed(existing_tables):
                        echo_info(f"Dropping existing table: {existing_table}")
                        self.conn.execute(text(f"DROP TABLE \"{existing_table}\""))

            # Finally, create each table and its indexes/constraints
            for table_name, sql_chunk in tbl_chunks.items():
                if missing_tables_only and table_name in existing_tables: continue
                echo_info(f"Creating table: {table_name}")
                self.conn.execute(text(sql_chunk))
                self.conn.commit()


    def warn_or_truncate(self, table_name, overwrite=False):
        res = self.conn.execute(text(f"""
            SELECT COUNT(*) FROM "{self.prefix}{table_name}"
            """)).fetchone()
        
        if res[0] == 0:
            return
        elif not overwrite:
            raise UsageError(f"Table {self.prefix}{table_name} contains rows. If "
                "you want to truncate and recreate this table, use --overwrite.")
        
        self.conn.execute(
            text(f"TRUNCATE \"{self.prefix}{table_name}\" RESTART IDENTITY"))

    #############################
    # BASIC FUNCTIONS FOR APPENDING TO, UPDATING, AND RETRIEVING FROM TABLES
    #############################

    def get_row_by_id(self, table_name, id):
        tbl = self.table(table_name)
        return self.conn.execute(tbl.select().where(tbl.c.id == id)).fetchone()

    def _gather_column_values_into_set(self, table_and_colnames):
        """Gather all distinct values in multiple columns into a set.
        This is a helper function that aids in avoiding ID collisions when inserting new rows."""
        existing_ids = set()
        for table_name, col_name in table_and_colnames:
            existing_ids_sql = text(f'SELECT DISTINCT "{col_name}" FROM "{table_name}"')
            existing_ids.update(pd.read_sql_query(existing_ids_sql, self.conn)[col_name])
        return existing_ids

    def append_df_to_table(self, df, table_name):
        df.to_sql(f"{self.prefix}{table_name}", self.conn, if_exists='append', 
            index=False)
        self.conn.commit()

    def update_table_from_df(self, df, table_name, key_col):
        tbl = self.table(table_name)
        for _, row in df.iterrows():
            key_val = row.pop(key_col)
            self.conn.execute(
                tbl.update()
                    .where(tbl.columns[key_col] == key_val)
                    .values(**row)
            )
        self.conn.commit()

    @contextmanager
    def foreign_keys_temporarily_dropped(self, table_name, fk_names=None):
        """Perform an action on the database with some foreign keys temporarily dropped.
        This is a context manager, so it's intended to be used in a `with` clause.
        
        IMPORTANT: This won't work with composite (multi-column) foreign keys."""
        tbl = self.table(table_name)
        fks = [fk for fk in tbl.foreign_key_constraints]
        if fk_names is not None:
            fks = [fk for fk in fks if fk.name in fk_names]
        
        for fk in fks:
            self.conn.execute(text(f"""
                ALTER TABLE "{self.prefix}{table_name}" 
                DROP CONSTRAINT "{fk.name}" 
                """))
        try:
            yield
        except:
            self.conn.rollback()
            raise
        finally:
            echo_info("Rebuilding foreign key constraints...")
            for fk in fks:
                self.conn.execute(text(f"""
                    ALTER TABLE "{self.prefix}{table_name}" 
                    ADD CONSTRAINT "{fk.name}" 
                    FOREIGN KEY ("{fk.columns[0].name}")
                    REFERENCES "{fk.referred_table.name}"("{fk.elements[0].column.name}")
                    ON DELETE {fk.ondelete}
                    ON UPDATE {fk.onupdate}
                    """))
                self.conn.commit()
            echo_info("Done.")


    #############################
    # EPISODES AND NOTES
    #############################

    def get_all_episodes(self):
        """Get all episodes that we will import notes for.
        Returns a pandas DataFrame with all rows."""
        sql_file = os.path.abspath(os.path.join(__file__, "../sql/episodes.sql"))
        with open(sql_file) as f:
            sql = self.interpolate_config_vars(f.read())
            eps_df = pd.read_sql_query(text(sql), self.conn)
        return eps_df

    def get_notes_without_text(self):
        """Get all notes and `NoteID`s that do not have corresponding `note_text`.
        Returns a pandas DataFrame with all rows."""
        notes_without_text_sql = text(f"""
            SELECT "id", "NoteID"
                FROM "{self.prefix}notes"
                WHERE "note_text" IS NULL
            """)
        return pd.read_sql_query(notes_without_text_sql, self.conn)

    def get_full_text_notes(self, note_ids):
        """Get all notes with `full_text` for a given iterable of `note`.`id`s.
        Returns a pandas DataFrame with all rows."""
        note_ids = [int(id) for id in note_ids]
        notes_with_full_text_sql = text(f"""
            SELECT "id", "note_text"
                FROM "{self.prefix}notes"
                WHERE "id" = ANY(:note_ids)
                ORDER BY "id" ASC
            """)
        return pd.read_sql_query(notes_with_full_text_sql, self.conn, 
            params={"note_ids": note_ids})
    
    def get_earliest_notes_with_feature(self, ep_ids, feat_strat, 
            time_limit=HUNDRED_YEARS_IN_SECONDS):
        """For the given episodes with `ep_ids` and a feature strategy `feat_strat`, 
        get the earliest non-empty note for each episode where that feature strategy has 
        successfully produced a `feature_value`.
        
        Returns a row including ..."""
        if feat_strat.task.output_type != out_t.FEATURE:
            raise RuntimeError("Only strategies for creating features are allowed.")

        params = {
            "ep_ids": [int(ep_id) for ep_id in ep_ids], 
            "feat_name": feat_strat.task.name,
            "strat_id": feat_strat.id,
            "time_limit": time_limit
        }
        select_earliest_notes_sql = text(f"""
            SELECT DISTINCT ON (n."FK_episode_id")
                    n."id", "feature_name", "feature_value",
                    e."id" AS "episode_id",
                    nf."id" AS "note_feature_id",
                    EXTRACT(EPOCH FROM ("DateOfServiceDTS" - "startDTS")) 
                        AS "relative_note_service_seconds"
                FROM "{self.prefix}notes" as n
                LEFT OUTER JOIN "{self.prefix}note_features" as nf
                    ON nf."FK_note_id" = n."id"
                        AND "feature_name" = :feat_name
                        AND "FK_strategy_id" = :strat_id
                LEFT JOIN "{self.ep_table}" as e
                    ON n."FK_episode_id" = e."id"
                WHERE "note_text" IS NOT NULL 
                    AND ("feature_name" = :feat_name AND "feature_value" IS NOT NULL)
                    AND n."FK_episode_id" = ANY(:ep_ids)
                    AND EXTRACT(EPOCH FROM ("DateOfServiceDTS" - "startDTS")) < :time_limit
                ORDER BY n."FK_episode_id",
                    "relative_note_service_seconds" ASC
            """)

        return pd.read_sql_query(select_earliest_notes_sql, self.conn, params=params)

    def get_notes_for_human_annotator(self, task, ep_or_pt_id):
        """Retrieves notes that would be relevant to a human creating annotations for a particular
        task regarding labeling of either episodes or patients. 
        
        Provide a `task` and specify the episode or patient with `ep_or_pt_id`.
        Returns a DataFrame."""
        if not isinstance(task, type) or not issubclass(task, AbstractTask):
            raise RuntimeError("First argument must be a task")
        
        if task.output_type == out_t.EPISODE_LABEL:
            relevant_notes_sql = text(f"""
                SELECT n."id", n."note_text"
                    FROM "{self.prefix}notes" AS n
                    LEFT JOIN "{self.ep_table}" as e
                        ON n."FK_episode_id" = e."id"
                    WHERE
                        n."FK_episode_id" = :ep_id
                        AND n."note_text" IS NOT NULL
                        AND {self.interpolate_config_vars(task.note_where_sql)}
                    ORDER BY "DateOfServiceDTS"
                """)
            df = pd.read_sql_query(relevant_notes_sql, self.conn, params={"ep_id": ep_or_pt_id})            
        else:
            raise NotImplementedError
        return df


    #############################
    # IMPORTING NOTES
    #############################

    def import_notes(self, notes):
        """Imports notes from a list containing the full text for each note.
        In the absence of any metadata, we generate reasonable default values.
        All notes are assumed to have a unique new episode they are associated with."""
        ids_to_collect = [(f"{self.prefix}notes", "NoteID"), (f"{self.prefix}notes", "PatientEncounterID"), 
            (self.ep_table, "admitPatientEncounterID"), (self.ep_table, "dcPatientEncounterID")]
        existing_ids = self._gather_column_values_into_set(ids_to_collect)
        current_timestamp = datetime.now()

        # Prepare data for bulk insert
        cohort_records = []
        episode_records = []
        note_records = []
        for note_text in notes:
            patient_encounter_id = note_id = generate_unique_id(existing_ids)
            patient_id = f"T{patient_encounter_id}"
            # FK_episode_id's will be generated after the episodes are inserted
            cohort_records.append({
                'infectionCriteria': True,
                'excl_ST0_combined': False
            })
            episode_records.append({
                'patientID': patient_id,
                'admitPatientEncounterID': patient_encounter_id,
                'dcPatientEncounterID': patient_encounter_id,
                'startDTS': current_timestamp
            })
            note_records.append({
                'NoteID': note_id,
                'PatientEncounterID': patient_encounter_id,
                'DateOfServiceDTS': current_timestamp,
                'note_text': note_text
            })

        # Insert episodes first to get their IDs for the FK_episode_id's
        episodes_df = pd.DataFrame(episode_records)
        episodes_df.to_sql(self.ep_table, self.conn, if_exists='append', index=False)
        self.conn.commit()

        # Get the newly inserted episode IDs
        # We need to query back the episodes we just inserted by their unique PatientEncounterID
        ep_ids_sql = text(f'''
            SELECT "id", "admitPatientEncounterID"
            FROM "{self.ep_table}"
            WHERE "admitPatientEncounterID" = ANY(:encounter_ids)
        ''')
        inserted_episodes = pd.read_sql_query(
            ep_ids_sql,
            self.conn,
            params={'encounter_ids': [rec['admitPatientEncounterID'] for rec in episode_records]}
        )

        # Create a mapping from PatientEncounterID to episode ID
        encounter_to_episode = dict(zip(
            inserted_episodes['admitPatientEncounterID'],
            inserted_episodes['id']
        ))

        # Now insert cohort records with the correct FK_episode_id
        for i, cohort_rec in enumerate(cohort_records):
            cohort_rec['FK_episode_id'] = encounter_to_episode[episode_records[i]['admitPatientEncounterID']]

        cohorts_df = pd.DataFrame(cohort_records)
        cohorts_df.to_sql(self.cohort_table, self.conn, if_exists='append', index=False)
        self.conn.commit()

        # Finally, insert notes with the correct FK_episode_id
        for i, note_rec in enumerate(note_records):
            note_rec['FK_episode_id'] = encounter_to_episode[note_rec['PatientEncounterID']]

        notes_df = pd.DataFrame(note_records)
        self.append_df_to_table(notes_df, 'notes')


    #############################
    # STRATEGIES AND TASKS
    #############################

    def get_or_register_task(self, task):
        """Gets the database row corresponding to a task (creating it if necessary)."""
        as_row = task.as_row()
        res = self.conn.execute(text(f"""
            SELECT * FROM "{self.prefix}tasks"
                WHERE "output_type" = :output_type
                AND "name" = :name
            """), as_row).fetchone()      
        if res is None:
            tbl = self.table("tasks")
            insert_res = self.conn.execute(
                tbl.insert().values(**as_row).returning(tbl.c.id)
            )
            self.conn.commit()
            res = self.get_row_by_id("tasks", insert_res.fetchone()[0])
        return res

    def get_or_register_strategy(self, strategy):
        """Gets the database row corresponding to a strategy (creating it if necessary)."""
        as_row = strategy.as_row()
        res = self.conn.execute(text(f"""
            SELECT * FROM "{self.prefix}strategies"
                WHERE "FK_task_id" = :FK_task_id
                AND "name" = :name
                AND "version" = :version
            """), as_row).fetchone()
        if res is None:
            tbl = self.table("strategies")
            insert_res = self.conn.execute(
                tbl.insert().values(**as_row).returning(tbl.c.id)
            )
            self.conn.commit()
            res = self.get_row_by_id("strategies", insert_res.fetchone()[0])
        return res

    def get_unfinished_ids_for_strategy_or_task(
            self, strategy_or_task, not_annot_by=None, having_notes=False, newer_than=None):
        """Gets the IDs for notes, episodes, or patients for which a particular strategy or
        task has not yet been completed.
        
        If `not_annot_by` is provided, IDs for entities that specifically have not been annotated 
        by the given username are returned.
        
        If `having_notes` is True, will further filter to only IDs for episodes/patients that 
        have at least one relevant note (specified by the task's `note_where_sql` attribute).
        
        If `newer_than` is provided, IDs"""

        strategy_sql_clause = having_notes_join_sql = having_notes_where_sql = ""
        if isinstance(strategy_or_task, type) and issubclass(strategy_or_task, AbstractTask):
            task = strategy_or_task
        elif isinstance(strategy_or_task, AbstractStrategy):
            strategy_sql_clause = f"AND \"FK_strategy_id\" = {strategy_or_task.id} "
            task = strategy_or_task.task
        else:
            raise RuntimeError("First argument must be a strategy or a task")

        human_annotator_sql_clause = "IS NULL"
        if not_annot_by is not None: human_annotator_sql_clause = f"= '{not_annot_by}'"
        if newer_than is not None:
            if not isinstance(newer_than, datetime): raise RuntimeError("newer_than not a datetime")
            if task.output_type not in LAST_UPDATED_COLUMNS: raise NotImplementedError
            last_updated_col = LAST_UPDATED_COLUMNS[task.output_type]
            strategy_sql_clause += f"AND \"{last_updated_col}\" >= '{newer_than}' "

        if having_notes:
            having_notes_join_sql = f"""
                LEFT OUTER JOIN "{self.prefix}notes" AS n
                    ON e."id" = n."FK_episode_id"
            """
            having_notes_where_sql = f"""
                AND ({self.interpolate_config_vars(task.note_where_sql)})
                AND n."note_text" IS NOT NULL
            """

        if task.output_type == out_t.SECTION:
            notes_without_section_sql = text(f"""
                SELECT n."id"
                    FROM "{self.prefix}notes" AS n
                    LEFT OUTER JOIN "{self.prefix}note_sections" as ns
                        ON ns."FK_note_id" = n."id"
                            AND "section_name" = '{task.name}'
                            {strategy_sql_clause}
                    WHERE "note_text" IS NOT NULL 
                        AND ("section_name" IS NULL OR 
                            ("section_name" = '{task.name}' AND "section_value" IS NULL))
                        AND ({self.interpolate_config_vars(task.note_where_sql)})
                    ORDER BY "id"
                """)
            df = pd.read_sql_query(notes_without_section_sql, self.conn)
        elif task.output_type == out_t.FEATURE:
            notes_without_feature_sql = text(f"""
                SELECT n."id"
                    FROM "{self.prefix}notes" AS n
                    LEFT OUTER JOIN "{self.prefix}note_features" as nf
                        ON nf."FK_note_id" = n."id"
                            AND "feature_name" = '{task.name}'
                            {strategy_sql_clause}
                    {self.interpolate_config_vars(task.note_join_sql)}
                    WHERE "note_text" IS NOT NULL 
                        AND ("feature_name" IS NULL OR 
                            ("feature_name" = '{task.name}' AND "feature_value" IS NULL))
                        AND ({self.interpolate_config_vars(task.note_where_sql)})
                    ORDER BY "id"
                """)
            df = pd.read_sql_query(notes_without_feature_sql, self.conn)
        elif task.output_type == out_t.EPISODE_LABEL:
            eps_without_label_sql = text(f"""
                SELECT DISTINCT e."id"
                    FROM "{self.ep_table}" AS e
                    LEFT OUTER JOIN "{self.prefix}episode_labels" as el
                        ON el."FK_episode_id" = e."id"
                            AND "task_name" = '{task.name}'
                            {strategy_sql_clause}
                            AND "FK_human_annotator" {human_annotator_sql_clause}
                    {self.interpolate_config_vars(task.episode_join_sql)}
                    {having_notes_join_sql}
                    WHERE el."task_name" IS NULL
                        AND ({self.interpolate_config_vars(task.episode_where_sql)})
                        {having_notes_where_sql}
                    ORDER BY "id"
                """)
            df = pd.read_sql_query(eps_without_label_sql, self.conn)            
        else:
            raise NotImplementedError  # We haven't yet implemented any PATIENT_LABEL tasks
        return df['id'].to_numpy()


    def get_ids_annotated_for_task(self, annot_task, id_task=None, annot_by=None):
        """Gets the IDs for notes, episodes, or patients (corresponding to task `id_task`) for 
        which a human has already provided annotations for task `annot_task`. These can be the 
        same task.
        
        If `annot_by` is provided, only IDs for entities annotated by the given human's
        username are included."""

        if id_task is None: id_task = annot_task
        annot_by_sql_clause = "IS NOT NULL"
        if annot_by is not None: annot_by_sql_clause = f"= '{annot_by}'"

        if id_task.output_type in (out_t.SECTION, out_t.FEATURE, out_t.NOTE_LABEL):
            if annot_task.output_type == out_t.EPISODE_LABEL:
                notes_annot_for_task_sql = text(f"""
                    SELECT DISTINCT n."id"
                        FROM "{self.prefix}notes" AS n
                        LEFT JOIN "{self.ep_table}" as e
                            ON n."FK_episode_id" = e."id"
                        LEFT OUTER JOIN "{self.prefix}episode_labels" as el
                            ON el."FK_episode_id" = e."id"
                                AND "task_name" = '{annot_task.name}'
                                AND "FK_human_annotator" {annot_by_sql_clause}
                        WHERE "task_name" IS NOT NULL
                        ORDER BY "id"
                    """)
                df = pd.read_sql_query(notes_annot_for_task_sql, self.conn)
            else:
                raise NotImplementedError
        elif (id_task.output_type == out_t.EPISODE_LABEL and 
                annot_task.output_type == out_t.EPISODE_LABEL):
            eps_annot_for_task_sql = text(f"""
                SELECT DISTINCT e."id"
                    FROM "{self.ep_table}" AS e
                    LEFT OUTER JOIN "{self.prefix}episode_labels" as el
                        ON el."FK_episode_id" = e."id"
                            AND "task_name" = '{annot_task.name}'
                            AND "FK_human_annotator" {annot_by_sql_clause}
                    WHERE "task_name" IS NOT NULL
                    ORDER BY "id"
                """)
            df = pd.read_sql_query(eps_annot_for_task_sql, self.conn)
        else:
            raise NotImplementedError  # We haven't yet implemented any PATIENT_LABEL tasks
        return df['id'].to_numpy()


    #############################
    # SECTIONS OF NOTES
    #############################

    def insert_note_section(self, note_id, strategy, section_value):
        if strategy.task.output_type != out_t.SECTION:
            raise RuntimeError("Only strategies for creating sections are allowed.")

        params = {
            "FK_note_id": note_id, 
            "section_name": strategy.task.name,
            "section_value": section_value,
            "FK_strategy_id": strategy.id,
            "now": datetime.now()
        }
        insert_section_sql = text(f"""
            INSERT INTO "{self.prefix}note_sections"
                    ("FK_note_id", "section_name", "section_value", "FK_strategy_id",
                    "last_updated")
                VALUES (:FK_note_id, :section_name, :section_value, :FK_strategy_id,
                    :now)
            """)
        
        self.conn.execute(insert_section_sql, params)
        self.conn.commit()
    
    def filter_to_notes_with_section(self, note_ids, strategy):
        """Takes an iterable of note_ids and returns only the ones that have at least one
        particular section extracted for the given strategy."""
        if strategy.task.output_type != out_t.SECTION:
            raise RuntimeError("Only strategies for creating sections are allowed.")
        
        select_note_ids_having_section_sql = text(f"""
            SELECT DISTINCT "FK_note_id"
                FROM "{self.prefix}note_sections"
                WHERE "section_name" = '{strategy.task.name}'
            """)
        df = pd.read_sql_query(select_note_ids_having_section_sql, self.conn) 
        note_ids_having_section = df['FK_note_id'].to_numpy()
        return np.intersect1d(note_ids, note_ids_having_section)

    
    def get_note_section(self, note_id, strategy):
        df = self.get_note_sections([note_id], strategy)
        return create_row_like(**df.iloc[0].to_dict())


    def get_note_sections(self, note_ids, strategy):
        if strategy.task.output_type != out_t.SECTION:
            raise RuntimeError("Only strategies for creating sections are allowed.")

        params = {
            "note_ids": [int(note_id) for note_id in note_ids],
            "section_name": strategy.task.name
        }
        select_sections_sql = text(f"""
            SELECT "id", "FK_note_id", "section_name", "section_value"
                FROM "{self.prefix}note_sections"
                WHERE "FK_note_id" = ANY(:note_ids)
                    AND "section_name" = :section_name
                ORDER BY "id" ASC
            """)

        return pd.read_sql_query(select_sections_sql, self.conn, params=params)



    #############################
    # FEATURES OF NOTES
    #############################

    def upsert_note_feature(self, note_id, strategy, llm_output_raw, feature_value=None, 
            note_section_id=None, runtime=None):
        if strategy.task.output_type != out_t.FEATURE:
            raise RuntimeError("Only strategies for creating features are allowed.")
        if runtime is None:
            try: runtime = strategy.stop_timer()
            except RuntimeError: runtime = None

        params = {
            "note_id": int(note_id), 
            "feature_name": strategy.task.name,
            "note_section_id": int(note_section_id), 
            "strategy_id": int(strategy.id), 
            "llm_output_raw": llm_output_raw, 
            "feature_value": feature_value,
            "start_time": datetime.now(),
            "strategy_runtime": runtime
        }
        update_nf_sql = text(f"""
            INSERT INTO "{self.prefix}note_features" ("FK_note_id", "feature_name",
                    "FK_strategy_id", "FK_note_section_id", "llm_output_raw", "feature_value",
                    "feature_updated", "strategy_runtime")
                VALUES (:note_id, :feature_name, :strategy_id, :note_section_id, 
                    :llm_output_raw, :feature_value, :start_time, :strategy_runtime)
                ON CONFLICT ("FK_note_id", "feature_name", "FK_strategy_id")
                DO UPDATE SET "FK_note_section_id" = :note_section_id, 
                    "llm_output_raw" = :llm_output_raw,
                    "feature_value" = :feature_value,
                    "feature_updated" = :start_time,
                    "strategy_runtime" = :strategy_runtime
            """)
        
        self.conn.execute(update_nf_sql, params)
        self.conn.commit()


    #############################
    # EPISODE LABELS
    #############################

    def get_episode_labels(self, ep_label_strategy, ep_ids=None, human_annotated=False):
        params = {
            "strategy_id": ep_label_strategy.id,
            "task_id": ep_label_strategy.task_id,
            "ep_ids": [int(ep_id) for ep_id in ep_ids] if ep_ids is not None else None
        }
        ep_ids_sql = '' if ep_ids is None else 'AND "FK_episode_id" = ANY(:ep_ids)'
        columns = ''' "id", "FK_note_feature_id", "FK_episode_id", "FK_strategy_id",
            "FK_task_id", "task_name", "label_name", "label_value", 
            "line_number", "FK_human_annotator"
            '''
        order_by = 'ORDER BY "FK_episode_id", "line_number", "label_name", "id"'

        if type(human_annotated) is str:
            params["human_annotator"] = human_annotated
            get_labels_sql = text(f"""
                SELECT {columns}
                    FROM "{self.prefix}episode_labels" 
                    WHERE "FK_task_id" = :task_id
                        AND "FK_strategy_id" IS NULL
                        AND "FK_human_annotator" = :human_annotator
                        {ep_ids_sql}
                    {order_by}
                """)        
        elif human_annotated is True:
            get_labels_sql = text(f"""
                SELECT {columns}
                    FROM "{self.prefix}episode_labels" 
                    WHERE "FK_task_id" = :task_id
                        AND "FK_strategy_id" IS NULL
                        AND "FK_human_annotator" IS NOT NULL
                        {ep_ids_sql}
                    {order_by}
                """)
        else:
            get_labels_sql = text(f"""
                SELECT {columns}
                    FROM "{self.prefix}episode_labels" 
                    WHERE "FK_task_id" = :task_id
                        AND "FK_strategy_id" = :strategy_id
                        AND "FK_human_annotator" IS NULL
                        {ep_ids_sql}
                    {order_by}
                """)

        return pd.read_sql_query(get_labels_sql, self.conn, params=params)


    def import_episode_labels(self, episode_label_task, input_xlsx, sheet_name=None, 
            human_username=None):
        if sheet_name is None: sheet_name = 0
        if human_username is None: human_username = getpass.getuser()
        self.ensure_annotator_exists(human_username)

        strats = find_strategies(output_type=out_t.EPISODE_LABEL, task=episode_label_task)
        if len(strats) == 0:
            raise RuntimeError("Could not find an episode labelling task by that name")
        strategy = strats[0](self, self.config)

        df = pd.read_excel(input_xlsx, sheet_name=sheet_name)
        long_df = df.assign(label_name=df['human_labels'].str.split(r'\s*[|]\s*'))
        long_df = long_df.explode('label_name').reset_index(drop=True)
        filtered_df = long_df.dropna(subset=['label_name'])
        if len(filtered_df) < len(long_df):
            echo_warn(f"{len(long_df) - len(filtered_df)} rows in the XLSX had zero labels")
        long_df = filtered_df[['FK_episode_id', 'label_name']].copy()

        vocab = strategy.task.vocab
        invalid_terms = [term for term in long_df.label_name if term not in vocab]
        if len(invalid_terms) > 0:
            raise RuntimeError(f"Invalid labels not in the vocab: {invalid_terms}")

        long_df['FK_task_id'] = strategy.task_id
        long_df['task_name'] = strategy.task.name
        long_df['FK_human_annotator'] = human_username
        long_df['label_value'] = 1.0

        params = {
            "ep_ids": [int(ep_id) for ep_id in long_df['FK_episode_id'].unique()],
            "task_id": strategy.task_id,
            "human_username": human_username
        }
        delete_labels_sql = text(f"""
            DELETE FROM "{self.prefix}episode_labels" 
                WHERE "FK_episode_id" = ANY(:ep_ids)
                    AND "FK_task_id" = :task_id
                    AND "FK_strategy_id" IS NULL
                    AND "FK_human_annotator" = :human_username
            """)
        self.conn.execute(delete_labels_sql, params)

        self.append_df_to_table(long_df, 'episode_labels')  # This also .commit()'s changes

        unique_eps = len(long_df['FK_episode_id'].unique())
        echo_info(f"{len(long_df)} labels imported for {unique_eps} episodes.")
    

    def replace_episode_labels(self, ep_label_strategy, row_values, labels_dict):
        if ep_label_strategy.task.output_type != out_t.EPISODE_LABEL:
            raise RuntimeError("Only strategies for creating episode labels are allowed.")
        
        params = {
            "episode_id": row_values.episode_id,
            "strategy_id": ep_label_strategy.id,
            "task_name": ep_label_strategy.task.name,
            "note_feature_id": row_values.note_feature_id,
            "task_id": ep_label_strategy.task_id
        }
        delete_labels_sql = text(f"""
            DELETE FROM "{self.prefix}episode_labels" 
                WHERE "FK_episode_id" = :episode_id
                    AND "FK_strategy_id" = :strategy_id
                    AND "task_name" = :task_name
                    AND "FK_human_annotator" IS NULL
            """)
        self.conn.execute(delete_labels_sql, params)

        for label_name, line_no in labels_dict.items():
            params.update({
                "label_name": label_name,
                "line_no": line_no
            })
            insert_label_sql = text(f"""
                INSERT INTO "{self.prefix}episode_labels" ("FK_note_feature_id", 
                    "FK_episode_id", "FK_strategy_id", "FK_task_id", "task_name", 
                    "label_name", "label_value", "line_number", "FK_human_annotator")
                VALUES (:note_feature_id, :episode_id, :strategy_id, :task_id, :task_name,
                    :label_name, 1.0, :line_no, NULL)
                """)
            self.conn.execute(insert_label_sql, params)
        
        self.conn.commit()


    #############################
    # HUMAN ANNOTATORS
    #############################
    def ensure_annotator_exists(self, human_username, warn=True):
        tbl = self.table("annotators")
        query = tbl.select().where(tbl.c.username == human_username)
        user = self.conn.execute(query).fetchone()
        if user is None:
            if warn: echo_warn(f"Creating entry for human annotator '{human_username}'")
            insert_user_sql = text(f"""
                INSERT INTO "{self.prefix}annotators" ("username", "admin")
                VALUES (:username, FALSE)
                """)
            self.conn.execute(insert_user_sql, {"username": human_username})


from .abstract import AbstractSectionTask

class ShortHPISectionTask(AbstractSectionTask):
    """\
    Simplest version of the HPI extractable from an H&P or ED note,
    starting with the summary statement and **not** including the ED course,
    A&P, any SmartText'ed ED data or findings, medications, A&P, ROS, 
    PMH, Impression, or other similar sections of the note.

    This may include the Onc history or descriptions of recent hospitalizations
    if they are included before the ED course or other sections as listed above.

    For now, we focus on H&Ps rather than the ED notes. We also don't consider
    Consult notes, although they'd probably work OK as well.
    """
    name = "hpi_short"
    note_where_sql = """
        "InpatientNoteTypeDSC" IN ('H&P')
    """